@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2016 Canonical Ltd
+ * Copyright (C) 2015-2018 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -66,6 +66,10 @@ type Config struct {
 
 	// Socket is the path to the unix socket to use
 	Socket string
+
+	// DisableKeepAlive indicates whether the connections should not be kept
+	// alive for later reuse
+	DisableKeepAlive bool
 }
 
 // A Client knows how to talk to the snappy daemon.
@@ -75,6 +79,11 @@ type Client struct {
 
 	disableAuth bool
 	interactive bool
+
+	maintenance error
+
+	warningCount     int
+	warningTimestamp time.Time
 }
 
 // New returns a new instance of Client
@@ -85,14 +94,13 @@ func New(config *Config) *Client {
 
 	// By default talk over an UNIX socket.
 	if config.BaseURL == "" {
+		transport := &http.Transport{Dial: unixDialer(config.Socket), DisableKeepAlives: config.DisableKeepAlive}
 		return &Client{
 			baseURL: url.URL{
 				Scheme: "http",
 				Host:   "localhost",
 			},
-			doer: &http.Client{
-				Transport: &http.Transport{Dial: unixDialer(config.Socket)},
-			},
+			doer:        &http.Client{Transport: transport},
 			disableAuth: config.DisableAuth,
 			interactive: config.Interactive,
 		}
@@ -104,12 +112,24 @@ func New(config *Config) *Client {
 	}
 	return &Client{
 		baseURL:     *baseURL,
-		doer:        &http.Client{},
+		doer:        &http.Client{Transport: &http.Transport{DisableKeepAlives: config.DisableKeepAlive}},
 		disableAuth: config.DisableAuth,
 		interactive: config.Interactive,
 	}
 }
 
+// Maintenance returns an error reflecting the daemon maintenance status or nil.
+func (client *Client) Maintenance() error {
+	return client.maintenance
+}
+
+// WarningsSummary returns the number of warnings that are ready to be shown to
+// the user, and the timestamp of the most recently added warning (useful for
+// silencing the warning alerts, and OKing the returned warnings).
+func (client *Client) WarningsSummary() (count int, timestamp time.Time) {
+	return client.warningCount, client.warningTimestamp
+}
+
 func (client *Client) WhoAmI() (string, error) {
 	user, err := readAuthData()
 	if os.IsNotExist(err) {
@@ -216,6 +236,19 @@ func MockDoRetry(retry, timeout time.Duration) (restore func()) {
 	}
 }
 
+type hijacked struct {
+	do func(*http.Request) (*http.Response, error)
+}
+
+func (h hijacked) Do(req *http.Request) (*http.Response, error) {
+	return h.do(req)
+}
+
+// Hijack lets the caller take over the raw http request
+func (client *Client) Hijack(f func(*http.Request) (*http.Response, error)) {
+	client.doer = hijacked{f}
+}
+
 // do performs a request and decodes the resulting json into the given
 // value. It's low-level, for testing/experimenting only; you should
 // usually use a higher level interface that builds on this.
@@ -243,29 +276,37 @@ func (client *Client) do(method, path string, query url.Values, headers map[stri
 	defer rsp.Body.Close()
 
 	if v != nil {
-		dec := json.NewDecoder(rsp.Body)
-		if err := dec.Decode(v); err != nil {
-			r := dec.Buffered()
-			buf, err1 := ioutil.ReadAll(r)
-			if err1 != nil {
-				buf = []byte(fmt.Sprintf("error reading buffered response body: %s", err1))
-			}
-			return fmt.Errorf("cannot decode %q: %s", buf, err)
+		if err := decodeInto(rsp.Body, v); err != nil {
+			return err
 		}
 	}
 
 	return nil
 }
 
+func decodeInto(reader io.Reader, v interface{}) error {
+	dec := json.NewDecoder(reader)
+	if err := dec.Decode(v); err != nil {
+		r := dec.Buffered()
+		buf, err1 := ioutil.ReadAll(r)
+		if err1 != nil {
+			buf = []byte(fmt.Sprintf("error reading buffered response body: %s", err1))
+		}
+		return fmt.Errorf("cannot decode %q: %s", buf, err)
+	}
+	return nil
+}
+
 // doSync performs a request to the given path using the specified HTTP method.
 // It expects a "sync" response from the API and on success decodes the JSON
-// response payload into the given value.
+// response payload into the given value using the "UseNumber" json decoding
+// which produces json.Numbers instead of float64 types for numbers.
 func (client *Client) doSync(method, path string, query url.Values, headers map[string]string, body io.Reader, v interface{}) (*ResultInfo, error) {
 	var rsp response
 	if err := client.do(method, path, query, headers, body, &rsp); err != nil {
 		return nil, err
 	}
-	if err := rsp.err(); err != nil {
+	if err := rsp.err(client); err != nil {
 		return nil, err
 	}
 	if rsp.Type != "sync" {
@@ -278,29 +319,37 @@ func (client *Client) doSync(method, path string, query url.Values, headers map[
 		}
 	}
 
+	client.warningCount = rsp.WarningCount
+	client.warningTimestamp = rsp.WarningTimestamp
+
 	return &rsp.ResultInfo, nil
 }
 
 func (client *Client) doAsync(method, path string, query url.Values, headers map[string]string, body io.Reader) (changeID string, err error) {
+	_, changeID, err = client.doAsyncFull(method, path, query, headers, body)
+	return
+}
+
+func (client *Client) doAsyncFull(method, path string, query url.Values, headers map[string]string, body io.Reader) (result json.RawMessage, changeID string, err error) {
 	var rsp response
 
 	if err := client.do(method, path, query, headers, body, &rsp); err != nil {
-		return "", err
+		return nil, "", err
 	}
-	if err := rsp.err(); err != nil {
-		return "", err
+	if err := rsp.err(client); err != nil {
+		return nil, "", err
 	}
 	if rsp.Type != "async" {
-		return "", fmt.Errorf("expected async response for %q on %q, got %q", method, path, rsp.Type)
+		return nil, "", fmt.Errorf("expected async response for %q on %q, got %q", method, path, rsp.Type)
 	}
 	if rsp.StatusCode != 202 {
-		return "", fmt.Errorf("operation not accepted")
+		return nil, "", fmt.Errorf("operation not accepted")
 	}
 	if rsp.Change == "" {
-		return "", fmt.Errorf("async response without change reference")
+		return nil, "", fmt.Errorf("async response without change reference")
 	}
 
-	return rsp.Change, nil
+	return rsp.Result, rsp.Change, nil
 }
 
 type ServerVersion struct {
@@ -339,7 +388,12 @@ type response struct {
 	Type       string          `json:"type"`
 	Change     string          `json:"change"`
 
+	WarningCount     int       `json:"warning-count"`
+	WarningTimestamp time.Time `json:"warning-timestamp"`
+
 	ResultInfo
+
+	Maintenance *Error `json:"maintenance"`
 }
 
 // Error is the real value of response.Result when an error occurs.
@@ -359,6 +413,7 @@ const (
 	ErrorKindTwoFactorRequired = "two-factor-required"
 	ErrorKindTwoFactorFailed   = "two-factor-failed"
 	ErrorKindLoginRequired     = "login-required"
+	ErrorKindInvalidAuthData   = "invalid-auth-data"
 	ErrorKindTermsNotAccepted  = "terms-not-accepted"
 	ErrorKindNoPaymentMethods  = "no-payment-methods"
 	ErrorKindPaymentDeclined   = "payment-declined"
@@ -367,15 +422,44 @@ const (
 	ErrorKindSnapAlreadyInstalled   = "snap-already-installed"
 	ErrorKindSnapNotInstalled       = "snap-not-installed"
 	ErrorKindSnapNotFound           = "snap-not-found"
+	ErrorKindAppNotFound            = "app-not-found"
 	ErrorKindSnapLocal              = "snap-local"
 	ErrorKindSnapNeedsDevMode       = "snap-needs-devmode"
 	ErrorKindSnapNeedsClassic       = "snap-needs-classic"
 	ErrorKindSnapNeedsClassicSystem = "snap-needs-classic-system"
+	ErrorKindSnapNotClassic         = "snap-not-classic"
 	ErrorKindNoUpdateAvailable      = "snap-no-update-available"
 
+	ErrorKindRevisionNotAvailable     = "snap-revision-not-available"
+	ErrorKindChannelNotAvailable      = "snap-channel-not-available"
+	ErrorKindArchitectureNotAvailable = "snap-architecture-not-available"
+
+	ErrorKindChangeConflict = "snap-change-conflict"
+
 	ErrorKindNotSnap = "snap-not-a-snap"
+
+	ErrorKindNetworkTimeout = "network-timeout"
+	ErrorKindDNSFailure     = "dns-failure"
+
+	ErrorKindInterfacesUnchanged = "interfaces-unchanged"
+
+	ErrorKindBadQuery           = "bad-query"
+	ErrorKindConfigNoSuchOption = "option-not-found"
+
+	ErrorKindSystemRestart = "system-restart"
+	ErrorKindDaemonRestart = "daemon-restart"
 )
 
+// IsRetryable returns true if the given error is an error
+// that can be retried later.
+func IsRetryable(err error) bool {
+	switch e := err.(type) {
+	case *Error:
+		return e.Kind == ErrorKindChangeConflict
+	}
+	return false
+}
+
 // IsTwoFactorError returns whether the given error is due to problems
 // in two-factor authentication.
 func IsTwoFactorError(err error) bool {
@@ -387,15 +471,30 @@ func IsTwoFactorError(err error) bool {
 	return e.Kind == ErrorKindTwoFactorFailed || e.Kind == ErrorKindTwoFactorRequired
 }
 
+// IsInterfacesUnchangedError returns whether the given error means the requested
+// change to interfaces was not made, because there was nothing to do.
+func IsInterfacesUnchangedError(err error) bool {
+	e, ok := err.(*Error)
+	if !ok || e == nil {
+		return false
+	}
+	return e.Kind == ErrorKindInterfacesUnchanged
+}
+
 // OSRelease contains information about the system extracted from /etc/os-release.
 type OSRelease struct {
 	ID        string `json:"id"`
 	VersionID string `json:"version-id,omitempty"`
 }
 
+// RefreshInfo contains information about refreshes.
 type RefreshInfo struct {
-	Schedule string `json:"schedule"`
+	// Timer contains the refresh.timer setting.
+	Timer string `json:"timer,omitempty"`
+	// Schedule contains the legacy refresh.schedule setting.
+	Schedule string `json:"schedule,omitempty"`
 	Last     string `json:"last,omitempty"`
+	Hold     string `json:"hold,omitempty"`
 	Next     string `json:"next,omitempty"`
 }
 
@@ -403,17 +502,28 @@ type RefreshInfo struct {
 type SysInfo struct {
 	Series    string    `json:"series,omitempty"`
 	Version   string    `json:"version,omitempty"`
+	BuildID   string    `json:"build-id"`
 	OSRelease OSRelease `json:"os-release"`
 	OnClassic bool      `json:"on-classic"`
 	Managed   bool      `json:"managed"`
 
 	KernelVersion string `json:"kernel-version,omitempty"`
 
-	Refresh     RefreshInfo `json:"refresh,omitempty"`
-	Confinement string      `json:"confinement"`
+	Refresh         RefreshInfo         `json:"refresh,omitempty"`
+	Confinement     string              `json:"confinement"`
+	SandboxFeatures map[string][]string `json:"sandbox-features,omitempty"`
 }
 
-func (rsp *response) err() error {
+func (rsp *response) err(cli *Client) error {
+	if cli != nil {
+		maintErr := rsp.Maintenance
+		// avoid setting to (*client.Error)(nil)
+		if maintErr != nil {
+			cli.maintenance = maintErr
+		} else {
+			cli.maintenance = nil
+		}
+	}
 	if rsp.Type != "error" {
 		return nil
 	}
@@ -438,7 +548,7 @@ func parseError(r *http.Response) error {
 		return fmt.Errorf("cannot unmarshal error: %v", err)
 	}
 
-	err := rsp.err()
+	err := rsp.err(nil)
 	if err == nil {
 		return fmt.Errorf("server error: %q", r.Status)
 	}
