@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2018 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,16 +20,17 @@
 package main
 
 import (
-	"bytes"
 	"fmt"
 	"io"
 	"path/filepath"
 	"strings"
 	"text/tabwriter"
-
-	"gopkg.in/yaml.v2"
+	"time"
+	"unicode"
+	"unicode/utf8"
 
 	"github.com/jessevdk/go-flags"
+	"gopkg.in/yaml.v2"
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/client"
@@ -40,15 +41,25 @@ import (
 )
 
 type infoCmd struct {
+	clientMixin
+	colorMixin
+	timeMixin
+
 	Verbose    bool `long:"verbose"`
 	Positional struct {
 		Snaps []anySnapName `positional-arg-name:"<snap>" required:"1"`
 	} `positional-args:"yes" required:"yes"`
 }
 
-var shortInfoHelp = i18n.G("show detailed information about a snap")
+var shortInfoHelp = i18n.G("Show detailed information about snaps")
 var longInfoHelp = i18n.G(`
-The info command shows detailed information about a snap, be it by name or by path.`)
+The info command shows detailed information about snaps.
+
+The snaps can be specified by name or by path; names are looked for both in the
+store and in the installed snaps; paths can refer to a .snap file, or to a
+directory that contains an unpacked snap suitable for 'snap try' (an example
+of this would be the 'prime' directory snapcraft produces).
+`)
 
 func init() {
 	addCommand("info",
@@ -56,9 +67,10 @@ func init() {
 		longInfoHelp,
 		func() flags.Commander {
 			return &infoCmd{}
-		}, map[string]string{
-			"verbose": i18n.G("Include a verbose list of a snap's notes (otherwise, summarise notes)"),
-		}, nil)
+		}, colorDescs.also(timeDescs).also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"verbose": i18n.G("Include more details on the snap (expanded notes, base, etc.)"),
+		}), nil)
 }
 
 func norm(path string) string {
@@ -100,6 +112,12 @@ func maybePrintID(w io.Writer, snap *client.Snap) {
 	}
 }
 
+func maybePrintBase(w io.Writer, base string, verbose bool) {
+	if verbose && base != "" {
+		fmt.Fprintf(w, "base:\t%s\n", base)
+	}
+}
+
 func tryDirect(w io.Writer, path string, verbose bool) bool {
 	path = norm(path)
 
@@ -122,7 +140,7 @@ func tryDirect(w io.Writer, path string, verbose bool) bool {
 		return false
 	}
 	fmt.Fprintf(w, "path:\t%q\n", path)
-	fmt.Fprintf(w, "name:\t%s\n", info.Name())
+	fmt.Fprintf(w, "name:\t%s\n", info.InstanceName())
 	fmt.Fprintf(w, "summary:\t%s\n", formatSummary(info.Summary()))
 
 	var notes *Notes
@@ -140,6 +158,7 @@ func tryDirect(w io.Writer, path string, verbose bool) bool {
 	}
 	fmt.Fprintf(w, "version:\t%s %s\n", info.Version, notes)
 	maybePrintType(w, string(info.Type))
+	maybePrintBase(w, info.Base, verbose)
 	if sha3_384 != "" {
 		fmt.Fprintf(w, "sha3-384:\t%s\n", sha3_384)
 	}
@@ -156,27 +175,90 @@ func coalesce(snaps ...*client.Snap) *client.Snap {
 	return nil
 }
 
-// formatDescr formats a given string (typically a snap description)
+// runesTrimRightSpace returns text, with any trailing whitespace dropped.
+func runesTrimRightSpace(text []rune) []rune {
+	j := len(text)
+	for j > 0 && unicode.IsSpace(text[j-1]) {
+		j--
+	}
+	return text[:j]
+}
+
+// runesLastIndexSpace returns the index of the last whitespace rune
+// in the text. If the text has no whitespace, returns -1.
+func runesLastIndexSpace(text []rune) int {
+	for i := len(text) - 1; i >= 0; i-- {
+		if unicode.IsSpace(text[i]) {
+			return i
+		}
+	}
+	return -1
+}
+
+// wrapLine wraps a line to fit into width, preserving the line's indent, and
+// writes it out prepending padding to each line.
+func wrapLine(out io.Writer, text []rune, pad string, width int) error {
+	// Note: this is _wrong_ for much of unicode (because the width of a rune on
+	//       the terminal is anything between 0 and 2, not always 1 as this code
+	//       assumes) but fixing that is Hard. Long story short, you can get close
+	//       using a couple of big unicode tables (which is what wcwidth
+	//       does). Getting it 100% requires a terminfo-alike of unicode behaviour.
+	//       However, before this we'd count bytes instead of runes, so we'd be
+	//       even more broken. Think of it as successive approximations... at least
+	//       with this work we share tabwriter's opinion on the width of things!
+
+	// This (and possibly printDescr below) should move to strutil once
+	// we're happy with it getting wider (heh heh) use.
+
+	// discard any trailing whitespace
+	text = runesTrimRightSpace(text)
+	// establish the indent of the whole block
+	idx := 0
+	for idx < len(text) && unicode.IsSpace(text[idx]) {
+		idx++
+	}
+	indent := pad + string(text[:idx])
+	text = text[idx:]
+	width -= idx + utf8.RuneCountInString(pad)
+	var err error
+	for len(text) > width && err == nil {
+		// find a good place to chop the text
+		idx = runesLastIndexSpace(text[:width+1])
+		if idx < 0 {
+			// there's no whitespace; just chop at line width
+			idx = width
+		}
+		_, err = fmt.Fprint(out, indent, string(text[:idx]), "\n")
+		// prune any remaining whitespace before the start of the next line
+		for idx < len(text) && unicode.IsSpace(text[idx]) {
+			idx++
+		}
+		text = text[idx:]
+	}
+	if err != nil {
+		return err
+	}
+	_, err = fmt.Fprint(out, indent, string(text), "\n")
+	return err
+}
+
+// printDescr formats a given string (typically a snap description)
 // in a user friendly way.
 //
 // The rules are (intentionally) very simple:
-// - trim whitespace
-// - word wrap at "max" chars
-// - keep \n intact and break here
-// - ignore \r
-func formatDescr(descr string, max int) string {
-	out := bytes.NewBuffer(nil)
-	for _, line := range strings.Split(strings.TrimSpace(descr), "\n") {
-		if len(line) > max {
-			for _, chunk := range strutil.WordWrap(line, max) {
-				fmt.Fprintf(out, "  %s\n", chunk)
-			}
-		} else {
-			fmt.Fprintf(out, "  %s\n", line)
+// - trim trailing whitespace
+// - word wrap at "max" chars preserving line indent
+// - keep \n intact and break there
+func printDescr(w io.Writer, descr string, max int) error {
+	var err error
+	descr = strings.TrimRightFunc(descr, unicode.IsSpace)
+	for _, line := range strings.Split(descr, "\n") {
+		err = wrapLine(w, []rune(line), "  ", max)
+		if err != nil {
+			break
 		}
 	}
-
-	return strings.TrimSuffix(out.String(), "\n")
+	return err
 }
 
 func maybePrintCommands(w io.Writer, snapName string, allApps []client.AppInfo, n int) {
@@ -237,37 +319,62 @@ func maybePrintServices(w io.Writer, snapName string, allApps []client.AppInfo,
 	}
 }
 
+var channelRisks = []string{"stable", "candidate", "beta", "edge"}
+
 // displayChannels displays channels and tracks in the right order
-func displayChannels(w io.Writer, remote *client.Snap) {
-	// \t\t\t so we get "installed" lined up with "channels"
-	fmt.Fprintf(w, "channels:\t\t\t\n")
+func (x *infoCmd) displayChannels(w io.Writer, chantpl string, esc *escapes, remote *client.Snap, revLen, sizeLen int) (maxRevLen, maxSizeLen int) {
+	fmt.Fprintln(w, "channels:")
+
+	releasedfmt := "2006-01-02"
+	if x.AbsTime {
+		releasedfmt = time.RFC3339
+	}
+
+	type chInfoT struct {
+		name, version, released, revision, size, notes string
+	}
+	var chInfos []*chInfoT
+	maxRevLen, maxSizeLen = revLen, sizeLen
 
 	// order by tracks
 	for _, tr := range remote.Tracks {
 		trackHasOpenChannel := false
-		for _, risk := range []string{"stable", "candidate", "beta", "edge"} {
+		for _, risk := range channelRisks {
 			chName := fmt.Sprintf("%s/%s", tr, risk)
 			ch, ok := remote.Channels[chName]
 			if tr == "latest" {
 				chName = risk
 			}
-			var version, revision, size, notes string
+			chInfo := chInfoT{name: chName}
 			if ok {
-				version = ch.Version
-				revision = fmt.Sprintf("(%s)", ch.Revision)
-				size = strutil.SizeToStr(ch.Size)
-				notes = NotesFromChannelSnapInfo(ch).String()
+				chInfo.version = ch.Version
+				chInfo.revision = fmt.Sprintf("(%s)", ch.Revision)
+				if len(chInfo.revision) > maxRevLen {
+					maxRevLen = len(chInfo.revision)
+				}
+				chInfo.released = ch.ReleasedAt.Format(releasedfmt)
+				chInfo.size = strutil.SizeToStr(ch.Size)
+				if len(chInfo.size) > maxSizeLen {
+					maxSizeLen = len(chInfo.size)
+				}
+				chInfo.notes = NotesFromChannelSnapInfo(ch).String()
 				trackHasOpenChannel = true
 			} else {
 				if trackHasOpenChannel {
-					version = "↑"
+					chInfo.version = esc.uparrow
 				} else {
-					version = "–" // that's an en dash (so yaml is happy)
+					chInfo.version = esc.dash
 				}
 			}
-			fmt.Fprintf(w, "  %s:\t%s\t%s\t%s\t%s\n", chName, version, revision, size, notes)
+			chInfos = append(chInfos, &chInfo)
 		}
 	}
+
+	for _, chInfo := range chInfos {
+		fmt.Fprintf(w, "  "+chantpl, chInfo.name, chInfo.version, chInfo.released, maxRevLen, chInfo.revision, maxSizeLen, chInfo.size, chInfo.notes)
+	}
+
+	return maxRevLen, maxSizeLen
 }
 
 func formatSummary(raw string) string {
@@ -279,8 +386,14 @@ func formatSummary(raw string) string {
 }
 
 func (x *infoCmd) Execute([]string) error {
-	cli := Client()
+	termWidth, _ := termSize()
+	termWidth -= 3
+	if termWidth > 100 {
+		// any wider than this and it gets hard to read
+		termWidth = 100
+	}
 
+	esc := x.getEscapes()
 	w := tabwriter.NewWriter(Stdout, 2, 2, 1, ' ', 0)
 
 	noneOK := true
@@ -289,36 +402,44 @@ func (x *infoCmd) Execute([]string) error {
 		if i > 0 {
 			fmt.Fprintln(w, "---")
 		}
+		if snapName == "system" {
+			fmt.Fprintln(w, "system: You can't have it.")
+			continue
+		}
 
 		if tryDirect(w, snapName, x.Verbose) {
 			noneOK = false
 			continue
 		}
-		remote, resInfo, _ := cli.FindOne(snapName)
-		local, _, _ := cli.Snap(snapName)
+		remote, resInfo, _ := x.client.FindOne(snapName)
+		local, _, _ := x.client.Snap(snapName)
 
 		both := coalesce(local, remote)
 
 		if both == nil {
-			fmt.Fprintf(w, "argument:\t%q\nwarning:\t%s\n", snapName, i18n.G("not a valid snap"))
+			if len(x.Positional.Snaps) == 1 {
+				return fmt.Errorf("no snap found for %q", snapName)
+			}
+
+			fmt.Fprintf(w, fmt.Sprintf(i18n.G("warning:\tno snap found for %q\n"), snapName))
 			continue
 		}
 		noneOK = false
 
 		fmt.Fprintf(w, "name:\t%s\n", both.Name)
 		fmt.Fprintf(w, "summary:\t%s\n", formatSummary(both.Summary))
-		// TODO: have publisher; use publisher here,
-		// and additionally print developer if publisher != developer
-		fmt.Fprintf(w, "publisher:\t%s\n", both.Developer)
+		fmt.Fprintf(w, "publisher:\t%s\n", longPublisher(esc, both.Publisher))
 		if both.Contact != "" {
 			fmt.Fprintf(w, "contact:\t%s\n", strings.TrimPrefix(both.Contact, "mailto:"))
 		}
+		license := both.License
+		if license == "" {
+			license = "unset"
+		}
+		fmt.Fprintf(w, "license:\t%s\n", license)
 		maybePrintPrice(w, remote, resInfo)
-		// FIXME: find out for real
-		termWidth := 77
-		fmt.Fprintf(w, "description: |\n%s\n", formatDescr(both.Description, termWidth))
-		maybePrintType(w, both.Type)
-		maybePrintID(w, both)
+		fmt.Fprintln(w, "description: |")
+		printDescr(w, both.Description, termWidth)
 		maybePrintCommands(w, snapName, both.Apps, termWidth)
 		maybePrintServices(w, snapName, both.Apps, termWidth)
 
@@ -328,8 +449,8 @@ func (x *infoCmd) Execute([]string) error {
 			fmt.Fprintf(w, "  confinement:\t%s\n", both.Confinement)
 		}
 
+		var notes *Notes
 		if local != nil {
-			var notes *Notes
 			if x.Verbose {
 				jailMode := local.Confinement == client.DevModeConfinement && !local.DevMode
 				fmt.Fprintf(w, "  devmode:\t%t\n", local.DevMode)
@@ -346,14 +467,37 @@ func (x *infoCmd) Execute([]string) error {
 			} else {
 				notes = NotesFromLocal(local)
 			}
-
-			fmt.Fprintf(w, "tracking:\t%s\n", local.TrackingChannel)
-			fmt.Fprintf(w, "installed:\t%s\t(%s)\t%s\t%s\n", local.Version, local.Revision, strutil.SizeToStr(local.InstalledSize), notes)
-			fmt.Fprintf(w, "refreshed:\t%s\n", local.InstallDate)
+		}
+		// stops the notes etc trying to be aligned with channels
+		w.Flush()
+		maybePrintType(w, both.Type)
+		maybePrintBase(w, both.Base, x.Verbose)
+		maybePrintID(w, both)
+		var localRev, localSize string
+		var revLen, sizeLen int
+		if local != nil {
+			if local.TrackingChannel != "" {
+				fmt.Fprintf(w, "tracking:\t%s\n", local.TrackingChannel)
+			}
+			if !local.InstallDate.IsZero() {
+				fmt.Fprintf(w, "refresh-date:\t%s\n", x.fmtTime(local.InstallDate))
+			}
+			localRev = fmt.Sprintf("(%s)", local.Revision)
+			revLen = len(localRev)
+			localSize = strutil.SizeToStr(local.InstalledSize)
+			sizeLen = len(localSize)
 		}
 
+		chantpl := "%s:\t%s %s%*s %*s %s\n"
 		if remote != nil && remote.Channels != nil && remote.Tracks != nil {
-			displayChannels(w, remote)
+			chantpl = "%s:\t%s\t%s\t%*s\t%*s\t%s\n"
+
+			w.Flush()
+			revLen, sizeLen = x.displayChannels(w, chantpl, esc, remote, revLen, sizeLen)
+		}
+		if local != nil {
+			fmt.Fprintf(w, chantpl,
+				"installed", local.Version, "", revLen, localRev, sizeLen, localSize, notes)
 		}
 
 	}
