@@ -23,153 +23,67 @@ import (
 	"errors"
 	"fmt"
 	"os"
-	"os/signal"
 	"path/filepath"
 	"sort"
 	"strings"
-	"syscall"
 	"time"
+	"unicode/utf8"
 
 	"github.com/jessevdk/go-flags"
 
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/osutil"
-	"github.com/snapcore/snapd/progress"
 )
 
-func lastLogStr(logs []string) string {
-	if len(logs) == 0 {
-		return ""
-	}
-	return logs[len(logs)-1]
-}
-
 var (
-	maxGoneTime = 5 * time.Second
-	pollTime    = 100 * time.Millisecond
-)
-
-type waitMixin struct {
-	NoWait bool `long:"no-wait" hidden:"true"`
-}
-
-// TODO: use waitMixin outside of cmd_snap_op.go
-
-var waitDescs = mixinDescs{
-	"no-wait": i18n.G("Do not wait for the operation to finish but just print the change id."),
-}
-
-var noWait = errors.New("no wait for op")
-
-func (wmx *waitMixin) wait(cli *client.Client, id string) (*client.Change, error) {
-	if wmx.NoWait {
-		fmt.Fprintf(Stdout, "%s\n", id)
-		return nil, noWait
-	}
-	return wait(cli, id)
-}
-
-func wait(cli *client.Client, id string) (*client.Change, error) {
-	pb := progress.MakeProgressBar()
-	defer func() {
-		pb.Finished()
-	}()
-
-	tMax := time.Time{}
-
-	var lastID string
-	lastLog := map[string]string{}
-	for {
-		chg, err := cli.Change(id)
-		if err != nil {
-			// a client.Error means we were able to communicate with
-			// the server (got an answer)
-			if e, ok := err.(*client.Error); ok {
-				return nil, e
-			}
-
-			// an non-client error here means the server most
-			// likely went away
-			// XXX: it actually can be a bunch of other things; fix client to expose it better
-			now := time.Now()
-			if tMax.IsZero() {
-				tMax = now.Add(maxGoneTime)
-			}
-			if now.After(tMax) {
-				return nil, err
-			}
-			pb.Spin(i18n.G("Waiting for server to restart"))
-			time.Sleep(pollTime)
-			continue
-		}
-		if !tMax.IsZero() {
-			pb.Finished()
-			tMax = time.Time{}
-		}
-
-		for _, t := range chg.Tasks {
-			switch {
-			case t.Status != "Doing":
-				continue
-			case t.Progress.Total == 1:
-				pb.Spin(t.Summary)
-				nowLog := lastLogStr(t.Log)
-				if lastLog[t.ID] != nowLog {
-					pb.Notify(nowLog)
-					lastLog[t.ID] = nowLog
-				}
-			case t.ID == lastID:
-				pb.Set(float64(t.Progress.Done))
-			default:
-				pb.Start(t.Summary, float64(t.Progress.Total))
-				lastID = t.ID
-			}
-			break
-		}
-
-		if chg.Ready {
-			if chg.Status == "Done" {
-				return chg, nil
-			}
-
-			if chg.Err != "" {
-				return chg, errors.New(chg.Err)
-			}
-
-			return nil, fmt.Errorf(i18n.G("change finished in status %q with no error message"), chg.Status)
-		}
-
-		// note this very purposely is not a ticker; we want
-		// to sleep 100ms between calls, not call once every
-		// 100ms.
-		time.Sleep(pollTime)
-	}
-}
-
-var (
-	shortInstallHelp = i18n.G("Installs a snap to the system")
-	shortRemoveHelp  = i18n.G("Removes a snap from the system")
-	shortRefreshHelp = i18n.G("Refreshes a snap in the system")
-	shortTryHelp     = i18n.G("Tests a snap in the system")
-	shortEnableHelp  = i18n.G("Enables a snap in the system")
-	shortDisableHelp = i18n.G("Disables a snap in the system")
+	shortInstallHelp = i18n.G("Install snaps on the system")
+	shortRemoveHelp  = i18n.G("Remove snaps from the system")
+	shortRefreshHelp = i18n.G("Refresh snaps in the system")
+	shortTryHelp     = i18n.G("Test an unpacked snap in the system")
+	shortEnableHelp  = i18n.G("Enable a snap in the system")
+	shortDisableHelp = i18n.G("Disable a snap in the system")
 )
 
 var longInstallHelp = i18n.G(`
-The install command installs the named snap in the system.
+The install command installs the named snaps on the system.
+
+To install multiple instances of the same snap, append an underscore and a
+unique identifier (for each instance) to a snap's name.
+
+With no further options, the snaps are installed tracking the stable channel,
+with strict security confinement.
+
+Revision choice via the --revision override requires the the user to
+have developer access to the snap, either directly or through the
+store's collaboration feature, and to be logged in (see 'snap help login').
+
+Note a later refresh will typically undo a revision override, taking the snap
+back to the current revision of the channel it's tracking.
+
+Use --name to set the instance name when installing from snap file.
 `)
 
 var longRemoveHelp = i18n.G(`
-The remove command removes the named snap from the system.
+The remove command removes the named snap instance from the system.
 
-By default all the snap revisions are removed, including their data and the common
-data directory. When a --revision option is passed only the specified revision is
-removed.
+By default all the snap revisions are removed, including their data and the
+common data directory. When a --revision option is passed only the specified
+revision is removed.
 `)
 
 var longRefreshHelp = i18n.G(`
-The refresh command refreshes (updates) the named snap.
+The refresh command updates the specified snaps, or all snaps in the system if
+none are specified.
+
+With no further options, the snaps are refreshed to the current revision of the
+channel they're tracking, preserving their confinement options.
+
+Revision choice via the --revision override requires the the user to
+have developer access to the snap, either directly or through the
+store's collaboration feature, and to be logged in (see 'snap help login').
+
+Note a later refresh will typically undo a revision override.
 `)
 
 var longTryHelp = i18n.G(`
@@ -189,7 +103,7 @@ The enable command enables a snap that was previously disabled.
 
 var longDisableHelp = i18n.G(`
 The disable command disables a snap. The binaries and services of the
-snap will no longer be available. But all the data is still available
+snap will no longer be available, but all the data is still available
 and the snap can easily be enabled again.
 `)
 
@@ -205,8 +119,7 @@ type cmdRemove struct {
 func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
 	name := string(x.Positional.Snaps[0])
 
-	cli := Client()
-	changeID, err := cli.Remove(name, opts)
+	changeID, err := x.client.Remove(name, opts)
 	if err != nil {
 		msg, err := errorToCmdMessage(name, err, opts)
 		if err != nil {
@@ -216,35 +129,33 @@ func (x *cmdRemove) removeOne(opts *client.SnapOptions) error {
 		return nil
 	}
 
-	_, err = x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
-	if err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
-	fmt.Fprintf(Stdout, i18n.G("%s removed\n"), name)
+	if opts.Revision != "" {
+		fmt.Fprintf(Stdout, i18n.G("%s (revision %s) removed\n"), name, opts.Revision)
+	} else {
+		fmt.Fprintf(Stdout, i18n.G("%s removed\n"), name)
+	}
 	return nil
 }
 
 func (x *cmdRemove) removeMany(opts *client.SnapOptions) error {
-	names := make([]string, len(x.Positional.Snaps))
-	for i, s := range x.Positional.Snaps {
-		names[i] = string(s)
-	}
-
-	cli := Client()
-	changeID, err := cli.RemoveMany(names, opts)
+	names := installedSnapNames(x.Positional.Snaps)
+	changeID, err := x.client.RemoveMany(names, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
+	chg, err := x.wait(changeID)
 	if err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -306,11 +217,16 @@ func (mxd mixinDescs) also(m map[string]string) mixinDescs {
 }
 
 var channelDescs = mixinDescs{
-	"channel":   i18n.G("Use this channel instead of stable"),
-	"beta":      i18n.G("Install from the beta channel"),
-	"edge":      i18n.G("Install from the edge channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"channel": i18n.G("Use this channel instead of stable"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"beta": i18n.G("Install from the beta channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"edge": i18n.G("Install from the edge channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
 	"candidate": i18n.G("Install from the candidate channel"),
-	"stable":    i18n.G("Install from the stable channel"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"stable": i18n.G("Install from the stable channel"),
 }
 
 func (mx *channelMixin) setChannelFromCommandline() error {
@@ -342,8 +258,7 @@ func (mx *channelMixin) setChannelFromCommandline() error {
 }
 
 // show what has been done
-func showDone(names []string, op string) error {
-	cli := Client()
+func showDone(cli *client.Client, names []string, op string, opts *client.SnapOptions, esc *escapes) error {
 	snaps, err := cli.List(names, nil)
 	if err != nil {
 		return err
@@ -356,17 +271,25 @@ func showDone(names []string, op string) error {
 		}
 		switch op {
 		case "install":
-			if snap.Developer != "" {
-				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' installed")
-				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' installed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
+			if opts != nil && opts.Classic && snap.Confinement != client.ClassicConfinement {
+				// requested classic but the snap is not classic
+				head := i18n.G("Warning:")
+				// TRANSLATORS: the arg is a snap name (e.g. "some-snap")
+				warn := fill(fmt.Sprintf(i18n.G("flag --classic ignored for strictly confined snap %s"), snap.Name), utf8.RuneCountInString(head)+1) // +1 for the space
+				fmt.Fprint(Stderr, head, " ", warn, "\n\n")
+			}
+
+			if snap.Publisher != nil {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from Alice installed")
+				fmt.Fprintf(Stdout, i18n.G("%s%s %s from %s installed\n"), snap.Name, channelStr, snap.Version, longPublisher(esc, snap.Publisher))
 			} else {
 				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 installed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s installed\n"), snap.Name, channelStr, snap.Version)
 			}
 		case "refresh":
-			if snap.Developer != "" {
-				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from 'alice' refreshed")
-				fmt.Fprintf(Stdout, i18n.G("%s%s %s from '%s' refreshed\n"), snap.Name, channelStr, snap.Version, snap.Developer)
+			if snap.Publisher != nil {
+				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version, then the developer name (e.g. "some-snap (beta) 1.3 from Alice refreshed")
+				fmt.Fprintf(Stdout, i18n.G("%s%s %s from %s refreshed\n"), snap.Name, channelStr, snap.Version, longPublisher(esc, snap.Publisher))
 			} else {
 				// TRANSLATORS: the args are a snap name optionally followed by a channel, then a version (e.g. "some-snap (beta) 1.3 refreshed")
 				fmt.Fprintf(Stdout, i18n.G("%s%s %s refreshed\n"), snap.Name, channelStr, snap.Version)
@@ -375,11 +298,11 @@ func showDone(names []string, op string) error {
 			// TRANSLATORS: first %s is a snap name, second %s is a revision
 			fmt.Fprintf(Stdout, i18n.G("%s reverted to %s\n"), snap.Name, snap.Version)
 		default:
-			fmt.Fprintf(Stdout, "internal error, unknown op %q", op)
+			fmt.Fprintf(Stdout, "internal error: unknown op %q", op)
 		}
-		if snap.TrackingChannel != snap.Channel {
-			// TRANSLATORS: first %s is a snap name, following %s is a channel name
-			fmt.Fprintf(Stdout, i18n.G("This leaves %s tracking %s.\n"), snap.Name, snap.TrackingChannel)
+		if snap.TrackingChannel != snap.Channel && snap.Channel != "" {
+			// TRANSLATORS: first %s is a channel name, following %s is a snap name, last %s is a channel name again.
+			fmt.Fprintf(Stdout, i18n.G("Channel %s for %s is closed; temporarily forwarding to %s.\n"), snap.TrackingChannel, snap.Name, snap.Channel)
 		}
 	}
 
@@ -397,8 +320,11 @@ type modeMixin struct {
 }
 
 var modeDescs = mixinDescs{
-	"classic":  i18n.G("Put snap in classic mode and disable security confinement"),
-	"devmode":  i18n.G("Put snap in development mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"classic": i18n.G("Put snap in classic mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
+	"devmode": i18n.G("Put snap in development mode and disable security confinement"),
+	// TRANSLATORS: This should not start with a lowercase letter.
 	"jailmode": i18n.G("Put snap in enforced confinement mode"),
 }
 
@@ -422,6 +348,7 @@ func (mx modeMixin) setModes(opts *client.SnapOptions) {
 }
 
 type cmdInstall struct {
+	colorMixin
 	waitMixin
 
 	channelMixin
@@ -435,39 +362,31 @@ type cmdInstall struct {
 
 	Unaliased bool `long:"unaliased"`
 
+	Name string `long:"name"`
+
 	Positional struct {
 		Snaps []remoteSnapName `positional-arg-name:"<snap>"`
 	} `positional-args:"yes" required:"yes"`
 }
 
-func setupAbortHandler(changeId string) {
-	// Intercept sigint
-	c := make(chan os.Signal, 2)
-	signal.Notify(c, syscall.SIGINT)
-	go func() {
-		<-c
-		cli := Client()
-		_, err := cli.Abort(changeId)
-		if err != nil {
-			fmt.Fprintf(Stderr, err.Error()+"\n")
-		}
-	}()
-}
-
-func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
+func (x *cmdInstall) installOne(nameOrPath, desiredName string, opts *client.SnapOptions) error {
 	var err error
-	var installFromFile bool
 	var changeID string
+	var snapName string
+	var path string
 
-	cli := Client()
-	if strings.Contains(name, "/") || strings.HasSuffix(name, ".snap") || strings.Contains(name, ".snap.") {
-		installFromFile = true
-		changeID, err = cli.InstallPath(name, opts)
+	if strings.Contains(nameOrPath, "/") || strings.HasSuffix(nameOrPath, ".snap") || strings.Contains(nameOrPath, ".snap.") {
+		path = nameOrPath
+		changeID, err = x.client.InstallPath(path, x.Name, opts)
 	} else {
-		changeID, err = cli.Install(name, opts)
+		snapName = nameOrPath
+		if desiredName != "" {
+			return errors.New(i18n.G("cannot use explicit name when installing from store"))
+		}
+		changeID, err = x.client.Install(snapName, opts)
 	}
 	if err != nil {
-		msg, err := errorToCmdMessage(name, err, opts)
+		msg, err := errorToCmdMessage(nameOrPath, err, opts)
 		if err != nil {
 			return err
 		}
@@ -475,27 +394,22 @@ func (x *cmdInstall) installOne(name string, opts *client.SnapOptions) error {
 		return nil
 	}
 
-	setupAbortHandler(changeID)
-
-	chg, err := x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
+	chg, err := x.wait(changeID)
 	if err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
 	// extract the snapName from the change, important for sideloaded
-	var snapName string
-
-	if installFromFile {
+	if path != "" {
 		if err := chg.Get("snap-name", &snapName); err != nil {
-			return fmt.Errorf("cannot extract the snap-name from local file %q: %s", name, err)
+			return fmt.Errorf("cannot extract the snap-name from local file %q: %s", nameOrPath, err)
 		}
-		name = snapName
 	}
 
-	return showDone([]string{name}, "install")
+	return showDone(x.client, []string{snapName}, "install", opts, x.getEscapes())
 }
 
 func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error {
@@ -506,19 +420,25 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 		}
 	}
 
-	cli := Client()
-	changeID, err := cli.InstallMany(names, opts)
+	changeID, err := x.client.InstallMany(names, opts)
 	if err != nil {
-		return err
-	}
-
-	setupAbortHandler(changeID)
-
-	chg, err := x.wait(cli, changeID)
-	if err == noWait {
+		var snapName string
+		if err, ok := err.(*client.Error); ok {
+			snapName, _ = err.Value.(string)
+		}
+		msg, err := errorToCmdMessage(snapName, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
 		return nil
 	}
+
+	chg, err := x.wait(changeID)
 	if err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -528,7 +448,7 @@ func (x *cmdInstall) installMany(names []string, opts *client.SnapOptions) error
 	}
 
 	if len(installed) > 0 {
-		if err := showDone(installed, "install"); err != nil {
+		if err := showDone(x.client, installed, "install", opts, x.getEscapes()); err != nil {
 			return err
 		}
 	}
@@ -566,28 +486,38 @@ func (x *cmdInstall) Execute([]string) error {
 	}
 	x.setModes(opts)
 
-	names := make([]string, len(x.Positional.Snaps))
-	for i, name := range x.Positional.Snaps {
-		names[i] = string(name)
+	names := remoteSnapNames(x.Positional.Snaps)
+	if len(names) == 0 {
+		return errors.New(i18n.G("cannot install zero snaps"))
+	}
+	for _, name := range names {
+		if len(name) == 0 {
+			return errors.New(i18n.G("cannot install snap with empty name"))
+		}
 	}
 
 	if len(names) == 1 {
-		return x.installOne(names[0], opts)
+		return x.installOne(names[0], x.Name, opts)
 	}
 
 	if x.asksForMode() || x.asksForChannel() {
 		return errors.New(i18n.G("a single snap name is needed to specify mode or channel flags"))
 	}
 
+	if x.Name != "" {
+		return errors.New(i18n.G("cannot use instance name when installing multiple snaps"))
+	}
 	return x.installMany(names, nil)
 }
 
 type cmdRefresh struct {
+	colorMixin
+	timeMixin
 	waitMixin
-
 	channelMixin
 	modeMixin
 
+	Amend            bool   `long:"amend"`
 	Revision         string `long:"revision"`
 	List             bool   `long:"list"`
 	Time             bool   `long:"time"`
@@ -598,17 +528,16 @@ type cmdRefresh struct {
 }
 
 func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error {
-	cli := Client()
-	changeID, err := cli.RefreshMany(snaps, opts)
+	changeID, err := x.client.RefreshMany(snaps, opts)
 	if err != nil {
 		return err
 	}
 
-	chg, err := x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
+	chg, err := x.wait(changeID)
 	if err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -618,7 +547,7 @@ func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error
 	}
 
 	if len(refreshed) > 0 {
-		return showDone(refreshed, "refresh")
+		return showDone(x.client, refreshed, "refresh", opts, x.getEscapes())
 	}
 
 	fmt.Fprintln(Stderr, i18n.G("All snaps up to date."))
@@ -627,8 +556,7 @@ func (x *cmdRefresh) refreshMany(snaps []string, opts *client.SnapOptions) error
 }
 
 func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
-	cli := Client()
-	changeID, err := cli.Refresh(name, opts)
+	changeID, err := x.client.Refresh(name, opts)
 	if err != nil {
 		msg, err := errorToCmdMessage(name, err, opts)
 		if err != nil {
@@ -638,32 +566,58 @@ func (x *cmdRefresh) refreshOne(name string, opts *client.SnapOptions) error {
 		return nil
 	}
 
-	_, err = x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
-	if err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
-	return showDone([]string{name}, "refresh")
+	return showDone(x.client, []string{name}, "refresh", opts, x.getEscapes())
+}
+
+func parseSysinfoTime(s string) time.Time {
+	t, err := time.Parse(time.RFC3339, s)
+	if err != nil {
+		return time.Time{}
+	}
+	return t
 }
 
 func (x *cmdRefresh) showRefreshTimes() error {
-	cli := Client()
-	sysinfo, err := cli.SysInfo()
+	sysinfo, err := x.client.SysInfo()
 	if err != nil {
 		return err
 	}
 
-	fmt.Fprintf(Stdout, "schedule: %s\n", sysinfo.Refresh.Schedule)
-	if sysinfo.Refresh.Last != "" {
-		fmt.Fprintf(Stdout, "last: %s\n", sysinfo.Refresh.Last)
+	if sysinfo.Refresh.Timer != "" {
+		fmt.Fprintf(Stdout, "timer: %s\n", sysinfo.Refresh.Timer)
+	} else if sysinfo.Refresh.Schedule != "" {
+		fmt.Fprintf(Stdout, "schedule: %s\n", sysinfo.Refresh.Schedule)
+	} else {
+		return errors.New("internal error: both refresh.timer and refresh.schedule are empty")
+	}
+	last := parseSysinfoTime(sysinfo.Refresh.Last)
+	hold := parseSysinfoTime(sysinfo.Refresh.Hold)
+	next := parseSysinfoTime(sysinfo.Refresh.Next)
+
+	if !last.IsZero() {
+		fmt.Fprintf(Stdout, "last: %s\n", x.fmtTime(last))
 	} else {
 		fmt.Fprintf(Stdout, "last: n/a\n")
 	}
-	if sysinfo.Refresh.Next != "" {
-		fmt.Fprintf(Stdout, "next: %s\n", sysinfo.Refresh.Next)
+	if !hold.IsZero() {
+		fmt.Fprintf(Stdout, "hold: %s\n", x.fmtTime(hold))
+	}
+	// only show "next" if its after "hold" to not confuse users
+	if !next.IsZero() {
+		// Snapstate checks for holdTime.After(limitTime) so we need
+		// to check for before or equal here to be fully correct.
+		if next.Before(hold) || next.Equal(hold) {
+			fmt.Fprintf(Stdout, "next: %s (but held)\n", x.fmtTime(next))
+		} else {
+			fmt.Fprintf(Stdout, "next: %s\n", x.fmtTime(next))
+		}
 	} else {
 		fmt.Fprintf(Stdout, "next: n/a\n")
 	}
@@ -671,8 +625,7 @@ func (x *cmdRefresh) showRefreshTimes() error {
 }
 
 func (x *cmdRefresh) listRefresh() error {
-	cli := Client()
-	snaps, _, err := cli.Find(&client.FindOptions{
+	snaps, _, err := x.client.Find(&client.FindOptions{
 		Refresh: true,
 	})
 	if err != nil {
@@ -685,12 +638,14 @@ func (x *cmdRefresh) listRefresh() error {
 
 	sort.Sort(snapsByName(snaps))
 
+	esc := x.getEscapes()
 	w := tabWriter()
 	defer w.Flush()
 
-	fmt.Fprintln(w, i18n.G("Name\tVersion\tRev\tDeveloper\tNotes"))
+	// TRANSLATORS: the %s is to insert a filler escape sequence (please keep it flush to the column header, with no extra spaces)
+	fmt.Fprintf(w, i18n.G("Name\tVersion\tRev\tPublisher%s\tNotes\n"), fillerPublisher(esc))
 	for _, snap := range snaps {
-		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, snap.Developer, NotesFromRemote(snap, nil))
+		fmt.Fprintf(w, "%s\t%s\t%s\t%s\t%s\n", snap.Name, snap.Version, snap.Revision, shortPublisher(esc, snap.Publisher), NotesFromRemote(snap, nil))
 	}
 
 	return nil
@@ -706,15 +661,14 @@ func (x *cmdRefresh) Execute([]string) error {
 
 	if x.Time {
 		if x.asksForMode() || x.asksForChannel() {
-			return errors.New(i18n.G("--time does not take mode nor channel flags"))
+			return errors.New(i18n.G("--time does not take mode or channel flags"))
 		}
-
 		return x.showRefreshTimes()
 	}
 
 	if x.List {
-		if x.asksForMode() || x.asksForChannel() {
-			return errors.New(i18n.G("--list does not take mode nor channel flags"))
+		if len(x.Positional.Snaps) > 0 || x.asksForMode() || x.asksForChannel() {
+			return errors.New(i18n.G("--list does not accept additional arguments"))
 		}
 
 		return x.listRefresh()
@@ -725,12 +679,10 @@ func (x *cmdRefresh) Execute([]string) error {
 		return nil
 	}
 
-	names := make([]string, len(x.Positional.Snaps))
-	for i, name := range x.Positional.Snaps {
-		names[i] = string(name)
-	}
-	if len(x.Positional.Snaps) == 1 {
+	names := installedSnapNames(x.Positional.Snaps)
+	if len(names) == 1 {
 		opts := &client.SnapOptions{
+			Amend:            x.Amend,
 			Channel:          x.Channel,
 			IgnoreValidation: x.IgnoreValidation,
 			Revision:         x.Revision,
@@ -777,7 +729,6 @@ func (x *cmdTry) Execute([]string) error {
 	if err := x.validateMode(); err != nil {
 		return err
 	}
-	cli := Client()
 	name := x.Positional.SnapDir
 	opts := &client.SnapOptions{}
 	x.setModes(opts)
@@ -801,21 +752,21 @@ func (x *cmdTry) Execute([]string) error {
 		return fmt.Errorf(i18n.G("cannot get full path for %q: %v"), name, err)
 	}
 
-	changeID, err := cli.Try(path, opts)
-	if e, ok := err.(*client.Error); ok && e.Kind == client.ErrorKindNotSnap {
-		return fmt.Errorf(i18n.G(`%q does not contain an unpacked snap.
-
-Try "snapcraft prime" in your project directory, then "snap try" again.`), path)
-	}
+	changeID, err := x.client.Try(path, opts)
 	if err != nil {
-		return err
-	}
-
-	chg, err := x.wait(cli, changeID)
-	if err == noWait {
+		msg, err := errorToCmdMessage(name, err, opts)
+		if err != nil {
+			return err
+		}
+		fmt.Fprintln(Stderr, msg)
 		return nil
 	}
+
+	chg, err := x.wait(changeID)
 	if err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -828,7 +779,7 @@ Try "snapcraft prime" in your project directory, then "snap try" again.`), path)
 	name = snapName
 
 	// show output as speced
-	snaps, err := cli.List([]string{name}, nil)
+	snaps, err := x.client.List([]string{name}, nil)
 	if err != nil {
 		return err
 	}
@@ -851,19 +802,17 @@ type cmdEnable struct {
 }
 
 func (x *cmdEnable) Execute([]string) error {
-	cli := Client()
 	name := string(x.Positional.Snap)
 	opts := &client.SnapOptions{}
-	changeID, err := cli.Enable(name, opts)
+	changeID, err := x.client.Enable(name, opts)
 	if err != nil {
 		return err
 	}
 
-	_, err = x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
-	if err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -880,19 +829,17 @@ type cmdDisable struct {
 }
 
 func (x *cmdDisable) Execute([]string) error {
-	cli := Client()
 	name := string(x.Positional.Snap)
 	opts := &client.SnapOptions{}
-	changeID, err := cli.Disable(name, opts)
+	changeID, err := x.client.Disable(name, opts)
 	if err != nil {
 		return err
 	}
 
-	_, err = x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
-	if err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -929,24 +876,22 @@ func (x *cmdRevert) Execute(args []string) error {
 		return err
 	}
 
-	cli := Client()
 	name := string(x.Positional.Snap)
 	opts := &client.SnapOptions{Revision: x.Revision}
 	x.setModes(opts)
-	changeID, err := cli.Revert(name, opts)
+	changeID, err := x.client.Revert(name, opts)
 	if err != nil {
 		return err
 	}
 
-	_, err = x.wait(cli, changeID)
-	if err == noWait {
-		return nil
-	}
-	if err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
-	return showDone([]string{name}, "revert")
+	return showDone(x.client, []string{name}, "revert", nil, nil)
 }
 
 var shortSwitchHelp = i18n.G("Switches snap to a different channel")
@@ -956,6 +901,7 @@ doing a refresh.
 `)
 
 type cmdSwitch struct {
+	waitMixin
 	channelMixin
 
 	Positional struct {
@@ -971,18 +917,20 @@ func (x cmdSwitch) Execute(args []string) error {
 		return fmt.Errorf("missing --channel=<channel-name> parameter")
 	}
 
-	cli := Client()
 	name := string(x.Positional.Snap)
 	channel := string(x.Channel)
 	opts := &client.SnapOptions{
 		Channel: channel,
 	}
-	changeID, err := cli.Switch(name, opts)
+	changeID, err := x.client.Switch(name, opts)
 	if err != nil {
 		return err
 	}
 
-	if _, err = wait(cli, changeID); err != nil {
+	if _, err := x.wait(changeID); err != nil {
+		if err == noWait {
+			return nil
+		}
 		return err
 	}
 
@@ -992,27 +940,42 @@ func (x cmdSwitch) Execute(args []string) error {
 
 func init() {
 	addCommand("remove", shortRemoveHelp, longRemoveHelp, func() flags.Commander { return &cmdRemove{} },
-		waitDescs.also(map[string]string{"revision": i18n.G("Remove only the given revision")}), nil)
+		waitDescs.also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Remove only the given revision"),
+		}), nil)
 	addCommand("install", shortInstallHelp, longInstallHelp, func() flags.Commander { return &cmdInstall{} },
-		waitDescs.also(channelDescs).also(modeDescs).also(map[string]string{
-			"revision":        i18n.G("Install the given revision of a snap, to which you must have developer access"),
-			"dangerous":       i18n.G("Install the given snap file even if there are no pre-acknowledged signatures for it, meaning it was not verified and could be dangerous (--devmode implies this)"),
+		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Install the given revision of a snap, to which you must have developer access"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"dangerous": i18n.G("Install the given snap file even if there are no pre-acknowledged signatures for it, meaning it was not verified and could be dangerous (--devmode implies this)"),
+			// TRANSLATORS: This should not start with a lowercase letter.
 			"force-dangerous": i18n.G("Alias for --dangerous (DEPRECATED)"),
-			"unaliased":       i18n.G("Install the given snap without enabling its automatic aliases"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"unaliased": i18n.G("Install the given snap without enabling its automatic aliases"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"name": i18n.G("Install the snap file under the given instance name"),
 		}), nil)
 	addCommand("refresh", shortRefreshHelp, longRefreshHelp, func() flags.Commander { return &cmdRefresh{} },
-		waitDescs.also(channelDescs).also(modeDescs).also(map[string]string{
-			"revision":          i18n.G("Refresh to the given revision"),
-			"list":              i18n.G("Show available snaps for refresh but do not perform a refresh"),
-			"time":              i18n.G("Show auto refresh information but do not perform a refresh"),
+		colorDescs.also(waitDescs).also(channelDescs).also(modeDescs).also(timeDescs).also(map[string]string{
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"amend": i18n.G("Allow refresh attempt on snap unknown to the store"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"revision": i18n.G("Refresh to the given revision, to which you must have developer access"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"list": i18n.G("Show the new versions of snaps that would be updated with the next refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"time": i18n.G("Show auto refresh information but do not perform a refresh"),
+			// TRANSLATORS: This should not start with a lowercase letter.
 			"ignore-validation": i18n.G("Ignore validation by other snaps blocking the refresh"),
 		}), nil)
 	addCommand("try", shortTryHelp, longTryHelp, func() flags.Commander { return &cmdTry{} }, waitDescs.also(modeDescs), nil)
 	addCommand("enable", shortEnableHelp, longEnableHelp, func() flags.Commander { return &cmdEnable{} }, waitDescs, nil)
 	addCommand("disable", shortDisableHelp, longDisableHelp, func() flags.Commander { return &cmdDisable{} }, waitDescs, nil)
 	addCommand("revert", shortRevertHelp, longRevertHelp, func() flags.Commander { return &cmdRevert{} }, waitDescs.also(modeDescs).also(map[string]string{
-		"revision": "Revert to the given revision",
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"revision": i18n.G("Revert to the given revision"),
 	}), nil)
-	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, nil, nil)
-
+	addCommand("switch", shortSwitchHelp, longSwitchHelp, func() flags.Commander { return &cmdSwitch{} }, waitDescs.also(channelDescs), nil)
 }
