@@ -32,9 +32,20 @@ var defaultTemplate = `
 
 #include <tunables/global>
 
+# snapd supports the concept of 'parallel installs' where snaps with the same
+# name are differentiated by '_<instance>' such that foo, foo_bar and foo_baz
+# may all be installed on the system. To support this, SNAP_NAME is set to the
+# name (eg, 'foo') while SNAP_INSTANCE_NAME is set to the instance name (eg
+# 'foo_bar'). The profile name and most rules therefore reference
+# SNAP_INSTANCE_NAME. In some cases, snapd will adjust the snap's runtime
+# environment so the snap doesn't have to be aware of the distinction (eg,
+# SNAP, SNAP_DATA and SNAP_COMMON are all bind mounted onto a directory with
+# SNAP_NAME so the security policy will allow writing to both locations (since
+# they are equivalent).
+
 ###VAR###
 
-###PROFILEATTACH### (attach_disconnected) {
+###PROFILEATTACH### (attach_disconnected,mediate_deleted) {
   #include <abstractions/base>
   #include <abstractions/consoles>
   #include <abstractions/openssl>
@@ -44,6 +55,7 @@ var defaultTemplate = `
   /etc/ld.so.preload r,
 
   # The base abstraction doesn't yet have this
+  /etc/sysconfig/clock r,
   /lib/terminfo/** rk,
   /usr/share/terminfo/** k,
   /usr/share/zoneinfo/** k,
@@ -57,8 +69,9 @@ var defaultTemplate = `
   # for details)
   deny /usr/lib/python3*/{,**/}__pycache__/ w,
   deny /usr/lib/python3*/{,**/}__pycache__/**.pyc.[0-9]* w,
-  deny @{INSTALL_DIR}/@{SNAP_NAME}/**/__pycache__/             w,
-  deny @{INSTALL_DIR}/@{SNAP_NAME}/**/__pycache__/*.pyc.[0-9]* w,
+  # bind mount used here (see 'parallel installs', above)
+  deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/             w,
+  deny @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**/__pycache__/*.pyc.[0-9]* w,
 
   # for perl apps/services
   #include <abstractions/perl>
@@ -92,8 +105,8 @@ var defaultTemplate = `
 
   # for bash 'binaries' (do *not* use abstractions/bash)
   # user-specific bash files
-  /bin/bash ixr,
-  /bin/dash ixr,
+  /{,usr/}bin/bash ixr,
+  /{,usr/}bin/dash ixr,
   /etc/bash.bashrc r,
   /etc/{passwd,group,nsswitch.conf} r,  # very common
   /etc/default/nss r,
@@ -106,6 +119,8 @@ var defaultTemplate = `
   # Common utilities for shell scripts
   /{,usr/}bin/arch ixr,
   /{,usr/}bin/{,g,m}awk ixr,
+  /{,usr/}bin/base32 ixr,
+  /{,usr/}bin/base64 ixr,
   /{,usr/}bin/basename ixr,
   /{,usr/}bin/bunzip2 ixr,
   /{,usr/}bin/bzcat ixr,
@@ -121,10 +136,14 @@ var defaultTemplate = `
   /{,usr/}bin/cpio ixr,
   /{,usr/}bin/cut ixr,
   /{,usr/}bin/date ixr,
+  /{,usr/}bin/dbus-daemon ixr,
+  /{,usr/}bin/dbus-run-session ixr,
+  /{,usr/}bin/dbus-send ixr,
   /{,usr/}bin/dd ixr,
   /{,usr/}bin/diff{,3} ixr,
   /{,usr/}bin/dir ixr,
   /{,usr/}bin/dirname ixr,
+  /{,usr/}bin/du ixr,
   /{,usr/}bin/echo ixr,
   /{,usr/}bin/{,e,f,r}grep ixr,
   /{,usr/}bin/env ixr,
@@ -133,6 +152,8 @@ var defaultTemplate = `
   /{,usr/}bin/find ixr,
   /{,usr/}bin/flock ixr,
   /{,usr/}bin/fmt ixr,
+  /{,usr/}bin/fold ixr,
+  /{,usr/}bin/getconf ixr,
   /{,usr/}bin/getent ixr,
   /{,usr/}bin/getopt ixr,
   /{,usr/}bin/groups ixr,
@@ -183,6 +204,7 @@ var defaultTemplate = `
   /{,usr/}bin/stat ixr,
   /{,usr/}bin/stdbuf ixr,
   /{,usr/}bin/stty ixr,
+  /{,usr/}bin/sync ixr,
   /{,usr/}bin/systemd-cat ixr,
   /{,usr/}bin/tac ixr,
   /{,usr/}bin/tail ixr,
@@ -215,6 +237,9 @@ var defaultTemplate = `
   # For snappy reexec on 4.8+ kernels
   /usr/lib/snapd/snap-exec m,
 
+  # For gdb support
+  /usr/lib/snapd/snap-gdb-shim ixr,
+
   # For in-snap tab completion
   /etc/bash_completion.d/{,*} r,
   /usr/lib/snapd/etelpmoc.sh ixr,               # marshaller (see complete.sh for out-of-snap unmarshal)
@@ -247,6 +272,7 @@ var defaultTemplate = `
 
   # snapctl and its requirements
   /usr/bin/snapctl ixr,
+  /usr/lib/snapd/snapctl ixr,
   @{PROC}/sys/net/core/somaxconn r,
   /run/snapd-snap.socket rw,
 
@@ -277,6 +303,15 @@ var defaultTemplate = `
   # value or those in its thread group.
   owner @{PROC}/@{pid}/task/@{tid}/comm rw,
 
+  # Allow reading and writing to our file descriptors in /proc which, for
+  # example, allow access to /dev/std{in,out,err} which are all symlinks to
+  # /proc/self/fd/{0,1,2} respectively. To support the open(..., O_TMPFILE)
+  # linkat() temporary file technique, allow all fds. Importantly, access to
+  # another's task's fd via this proc interface is mediated via 'ptrace (read)'
+  # (readonly) and 'ptrace (trace)' (read/write) which is denied by default, so
+  # this rule by itself doesn't allow opening another snap's fds via proc.
+  owner @{PROC}/@{pid}/{,task/@{tid}}fd/[0-9]* rw,
+
   # Miscellaneous accesses
   /dev/{,u}random w,
   /etc/machine-id r,
@@ -307,12 +342,14 @@ var defaultTemplate = `
   @{PROC}/sys/kernel/yama/ptrace_scope r,
   @{PROC}/sys/kernel/shmmax r,
   @{PROC}/sys/fs/file-max r,
+  @{PROC}/sys/fs/inotify/max_* r,
   @{PROC}/sys/kernel/pid_max r,
   @{PROC}/sys/kernel/random/uuid r,
   @{PROC}/sys/kernel/random/boot_id r,
   /sys/devices/virtual/tty/{console,tty*}/active r,
   /sys/fs/cgroup/memory/memory.limit_in_bytes r,
-  /sys/fs/cgroup/memory/snap.@{SNAP_NAME}{,.*}/memory.limit_in_bytes r,
+  /sys/fs/cgroup/memory/snap.@{SNAP_INSTANCE_NAME}{,.*}/memory.limit_in_bytes r,
+  /sys/module/apparmor/parameters/enabled r,
   /{,usr/}lib/ r,
 
   # Reads of oom_adj and oom_score_adj are safe
@@ -338,30 +375,38 @@ var defaultTemplate = `
   @{PROC}/net/dev r,
   @{PROC}/@{pid}/net/dev r,
 
+  # Read-only of this snap
+  /var/lib/snapd/snaps/@{SNAP_NAME}_*.snap r,
+
   # Read-only for the install directory
-  @{INSTALL_DIR}/@{SNAP_NAME}/                   r,
-  @{INSTALL_DIR}/@{SNAP_NAME}/@{SNAP_REVISION}/    r,
-  @{INSTALL_DIR}/@{SNAP_NAME}/@{SNAP_REVISION}/**  mrklix,
+  # bind mount used here (see 'parallel installs', above)
+  @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/                   r,
+  @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}/@{SNAP_REVISION}}/    r,
+  @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}/@{SNAP_REVISION}}/**  mrklix,
 
   # Read-only install directory for other revisions to help with bugs like
   # LP: #1616650 and LP: #1655992
-  @{INSTALL_DIR}/@{SNAP_NAME}/**  mrkix,
+  @{INSTALL_DIR}/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/**  mrkix,
 
   # Read-only home area for other versions
-  owner @{HOME}/snap/@{SNAP_NAME}/                  r,
-  owner @{HOME}/snap/@{SNAP_NAME}/**                mrkix,
+  # bind mount *not* used here (see 'parallel installs', above)
+  owner @{HOME}/snap/@{SNAP_INSTANCE_NAME}/                  r,
+  owner @{HOME}/snap/@{SNAP_INSTANCE_NAME}/**                mrkix,
 
   # Writable home area for this version.
-  owner @{HOME}/snap/@{SNAP_NAME}/@{SNAP_REVISION}/** wl,
-  owner @{HOME}/snap/@{SNAP_NAME}/common/** wl,
+  # bind mount *not* used here (see 'parallel installs', above)
+  owner @{HOME}/snap/@{SNAP_INSTANCE_NAME}/@{SNAP_REVISION}/** wl,
+  owner @{HOME}/snap/@{SNAP_INSTANCE_NAME}/common/** wl,
 
   # Read-only system area for other versions
-  /var/snap/@{SNAP_NAME}/   r,
-  /var/snap/@{SNAP_NAME}/** mrkix,
+  # bind mount used here (see 'parallel installs', above)
+  /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/   r,
+  /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/** mrkix,
 
   # Writable system area only for this version
-  /var/snap/@{SNAP_NAME}/@{SNAP_REVISION}/** wl,
-  /var/snap/@{SNAP_NAME}/common/** wl,
+  # bind mount used here (see 'parallel installs', above)
+  /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/@{SNAP_REVISION}/** wl,
+  /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/common/** wl,
 
   # The ubuntu-core-launcher creates an app-specific private restricted /tmp
   # and will fail to launch the app if something goes wrong. As such, we can
@@ -371,25 +416,35 @@ var defaultTemplate = `
 
   # App-specific access to files and directories in /dev/shm. We allow file
   # access in /dev/shm for shm_open() and files in subdirectories for open()
-  /{dev,run}/shm/snap.@{SNAP_NAME}.** mrwlkix,
+  # bind mount *not* used here (see 'parallel installs', above)
+  /{dev,run}/shm/snap.@{SNAP_INSTANCE_NAME}.** mrwlkix,
   # Also allow app-specific access for sem_open()
-  /{dev,run}/shm/sem.snap.@{SNAP_NAME}.* mrwk,
+  /{dev,run}/shm/sem.snap.@{SNAP_INSTANCE_NAME}.* mrwlk,
 
   # Snap-specific XDG_RUNTIME_DIR that is based on the UID of the user
-  owner /run/user/[0-9]*/snap.@{SNAP_NAME}/   rw,
-  owner /run/user/[0-9]*/snap.@{SNAP_NAME}/** mrwklix,
+  # bind mount *not* used here (see 'parallel installs', above)
+  owner /run/user/[0-9]*/snap.@{SNAP_INSTANCE_NAME}/   rw,
+  owner /run/user/[0-9]*/snap.@{SNAP_INSTANCE_NAME}/** mrwklix,
 
   # Allow apps from the same package to communicate with each other via an
   # abstract or anonymous socket
-  unix peer=(label=snap.@{SNAP_NAME}.*),
+  unix (bind, listen) addr="@snap.@{SNAP_INSTANCE_NAME}.**",
+  unix peer=(label=snap.@{SNAP_INSTANCE_NAME}.*),
 
   # Allow apps from the same package to communicate with each other via DBus.
   # Note: this does not grant access to the DBus sockets of well known buses
   # (will still need to use an appropriate interface for that).
-  dbus (receive, send) peer=(label=snap.@{SNAP_NAME}.*),
+  dbus (receive, send) peer=(label=snap.@{SNAP_INSTANCE_NAME}.*),
 
   # Allow apps from the same package to signal each other via signals
-  signal peer=snap.@{SNAP_NAME}.*,
+  signal peer=snap.@{SNAP_INSTANCE_NAME}.*,
+
+  # Allow receiving signals from all snaps (and focus on mediating sending of
+  # signals)
+  signal (receive) peer=snap.*,
+
+  # Allow receiving signals from unconfined (eg, systemd)
+  signal (receive) peer=unconfined,
 
   # for 'udevadm trigger --verbose --dry-run --tag-match=snappy-assign'
   /{,s}bin/udevadm ixr,
@@ -436,6 +491,11 @@ var defaultTemplate = `
   # Allow read-access to / for navigating to other parts of the filesystem.
   / r,
 
+  # Snap-specific run directory. Bind mount *not* used here
+  # (see 'parallel installs', above)
+  /run/snap.@{SNAP_INSTANCE_NAME}/ rw,
+  /run/snap.@{SNAP_INSTANCE_NAME}/** mrwklix,
+
 ###SNIPPETS###
 }
 `
@@ -454,7 +514,7 @@ var classicTemplate = `
 
 ###VAR###
 
-###PROFILEATTACH### (attach_disconnected) {
+###PROFILEATTACH### (attach_disconnected,mediate_deleted) {
   # set file rules so that exec() inherits our profile unless there is
   # already a profile for it (eg, snap-confine)
   / rwkl,
@@ -462,7 +522,7 @@ var classicTemplate = `
   /** pix,
 
   capability,
-  change_profile,
+  ###CHANGEPROFILE_RULE###
   dbus,
   network,
   mount,
@@ -487,6 +547,9 @@ var classicJailmodeSnippet = `
   # Read only access to the core snap to load libc from.
   # This is related to LP: #1666897
   @{INSTALL_DIR}/core/*/{,usr/}lib/@{multiarch}/{,**/}lib*.so* m,
+
+  # For snappy reexec on 4.8+ kernels
+  @{INSTALL_DIR}/core/*/usr/lib/snapd/snap-exec m,
 `
 
 // nfsSnippet contains extra permissions necessary for snaps and snap-confine
@@ -499,3 +562,166 @@ var nfsSnippet = `
   network inet,
   network inet6,
 `
+
+// overlayRootSnippet contains the extra permissions necessary for snap and
+// snap-confine to operate on systems where '/' is a writable overlay fs.
+// AppArmor requires directory reads for upperdir (but these aren't otherwise
+// visible to the snap). While we filter AppArmor regular expression (AARE)
+// characters elsewhere, we double quote the path in case UPPERDIR has spaces.
+var overlayRootSnippet = `
+  # snapd autogenerated workaround for systems using '/' on overlayfs. For
+  # details see: https://bugs.launchpad.net/apparmor/+bug/1703674
+  "###UPPERDIR###/{,**/}" r,
+`
+
+var ptraceTraceDenySnippet = `
+# While commands like 'ps', 'ip netns identify <pid>', 'ip netns pids foo', etc
+# trigger a 'ptrace (trace)' denial, they aren't actually tracing other
+# processes. Unfortunately, the kernel overloads trace such that the LSMs are
+# unable to distinguish between tracing other processes and other accesses.
+# ptrace (trace) can be used to break out of the seccomp sandbox unless the
+# kernel has 93e35efb8de45393cf61ed07f7b407629bf698ea (in 4.8+). Until snapd
+# has full ptrace support conditional on kernel support, explicitly deny to
+# silence noisy denials/avoid confusion and accidentally giving away this
+# dangerous access frivolously.
+deny ptrace (trace),
+`
+
+// updateNSTemplate defines the apparmor profile for per-snap snap-update-ns.
+//
+// The per-snap snap-update-ns profiles are composed via a template and
+// snippets for the snap. The template allows:
+// - accesses to libraries, files and /proc entries required to run
+// - using global and per-snap lock files
+// - reading per-snap mount namespaces and mount profiles
+// - managing per-snap freezer state files
+// - per-snap mounting/unmounting fonts from the host
+// - denying mounts to restricted places (eg, /snap/bin and /media)
+var updateNSTemplate = `
+# Description: Allows snap-update-ns to construct the mount namespace specific
+# to a particular snap (see the name below). This specifically includes the
+# precise locations of the layout elements.
+
+# vim:syntax=apparmor
+
+#include <tunables/global>
+
+profile snap-update-ns.###SNAP_INSTANCE_NAME### (attach_disconnected) {
+  # The next four rules mirror those above. We want to be able to read
+  # and map snap-update-ns into memory but it may come from a variety of places.
+  /usr/lib{,exec,64}/snapd/snap-update-ns mr,
+  /var/lib/snapd/hostfs/usr/lib{,exec,64}/snapd/snap-update-ns mr,
+  /{,var/lib/snapd/}snap/{core,snapd}/*/usr/lib/snapd/snap-update-ns mr,
+  /var/lib/snapd/hostfs/{,var/lib/snapd/}snap/core/*/usr/lib/snapd/snap-update-ns mr,
+
+  # Allow reading the dynamic linker cache.
+  /etc/ld.so.cache r,
+  # Allow reading, mapping and executing the dynamic linker.
+  /{,usr/}lib{,32,64,x32}/{,@{multiarch}/}ld-*.so mrix,
+  # Allow reading and mapping various parts of the standard library and
+  # dynamically loaded nss modules and what not.
+  /{,usr/}lib{,32,64,x32}/{,@{multiarch}/}libc{,-[0-9]*}.so* mr,
+  /{,usr/}lib{,32,64,x32}/{,@{multiarch}/}libpthread{,-[0-9]*}.so* mr,
+
+  # Common devices accesses
+  /dev/null rw,
+  /dev/full rw,
+  /dev/zero rw,
+  /dev/random r,
+  /dev/urandom r,
+
+  # Allow reading the command line (snap-update-ns uses it in pre-Go bootstrap code).
+  @{PROC}/@{pid}/cmdline r,
+
+  # Allow reading file descriptor paths
+  @{PROC}/@{pid}/fd/* r,
+  # Allow reading /proc/version. For release.go WSL detection.
+  @{PROC}/version r,
+
+  # Allow reading somaxconn, required in newer distro releases
+  @{PROC}/sys/net/core/somaxconn r,
+  # but silence noisy denial of inet/inet6
+  deny network inet,
+  deny network inet6,
+
+  # Allow reading the os-release file (possibly a symlink to /usr/lib).
+  /{etc/,usr/lib/}os-release r,
+
+  # Allow creating/grabbing global and per-snap lock files.
+  /run/snapd/lock/###SNAP_INSTANCE_NAME###.lock rwk,
+  /run/snapd/lock/.lock rwk,
+
+  # Allow reading stored mount namespaces,
+  /run/snapd/ns/ r,
+  /run/snapd/ns/###SNAP_INSTANCE_NAME###.mnt r,
+
+  # Allow reading per-snap desired mount profiles. Those are written by
+  # snapd and represent the desired layout and content connections.
+  /var/lib/snapd/mount/snap.###SNAP_INSTANCE_NAME###.fstab r,
+  /var/lib/snapd/mount/snap.###SNAP_INSTANCE_NAME###.user-fstab r,
+
+  # Allow reading and writing actual per-snap mount profiles. Note that
+  # the wildcard in the rule to allow an atomic write + rename strategy.
+  # Those files are written by snap-update-ns and represent the actual
+  # mount profile at a given moment.
+  /run/snapd/ns/snap.###SNAP_INSTANCE_NAME###.fstab{,.*} rw,
+
+  # NOTE: at this stage the /snap directory is stable as we have called
+  # pivot_root already.
+
+  # Needed to perform mount/unmounts.
+  capability sys_admin,
+  # Needed for mimic construction.
+  capability chown,
+  # Needed for dropping to calling user when processing per-user mounts
+  capability setuid,
+  capability setgid,
+  # Allow snap-update-ns to override file ownership and permission checks.
+  # This is required because writable mimics now preserve the permissions
+  # of the original and hence we may be asked to create a directory when the
+  # parent is a tmpfs without DAC write access.
+  capability dac_override,
+
+  # Allow freezing and thawing the per-snap cgroup freezers
+  /sys/fs/cgroup/freezer/snap.###SNAP_INSTANCE_NAME###/freezer.state rw,
+
+  # Allow the content interface to bind fonts from the host filesystem
+  mount options=(ro bind) /var/lib/snapd/hostfs/usr/share/fonts/ -> /snap/###SNAP_INSTANCE_NAME###/*/**,
+  umount /snap/###SNAP_INSTANCE_NAME###/*/**,
+
+  # set up user mount namespace
+  mount options=(rslave) -> /,
+
+  # Allow traversing from the root directory and several well-known places.
+  # Specific directory permissions are added by snippets below.
+  / r,
+  /etc/ r,
+  /snap/ r,
+  /tmp/ r,
+  /usr/ r,
+  /var/ r,
+  /var/snap/ r,
+
+  # Allow reading timezone data.
+  /usr/share/zoneinfo/** r,
+
+  # Don't allow anyone to touch /snap/bin
+  audit deny mount /snap/bin/** -> /**,
+  audit deny mount /** -> /snap/bin/**,
+
+  # Don't allow bind mounts to /media which has special
+  # sharing and propagates mount events outside of the snap namespace.
+  audit deny mount -> /media,
+
+  # Allow receiving signals from unconfined (eg, systemd)
+  signal (receive) peer=unconfined,
+  # Allow sending and receiving signals from ourselves.
+  signal peer=@{profile_name},
+
+  # Commonly needed permissions for writable mimics.
+  /tmp/ r,
+  /tmp/.snap/{,**} rw,
+
+###SNIPPETS###
+}
+`
