@@ -45,6 +45,10 @@ const avahiObserveBaseDeclarationSlots = `
 const avahiObservePermanentSlotAppArmor = `
 network netlink,
 
+# Allow access to daemon to create socket
+/{,var/}run/avahi-daemon/  w,
+/{,var/}run/avahi-daemon/{pid,socket} rw,
+
 # Description: Allow operating as the avahi service. This gives
 # privileged access to the system.
 #include <abstractions/dbus-strict>
@@ -116,7 +120,7 @@ dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.Server
     member=StateChanged
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # address resolving
 dbus (receive)
@@ -140,7 +144,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.HostNameResolver
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # service resolving
 dbus (receive)
@@ -152,7 +156,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.ServiceResolver
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # domain browsing
 dbus (receive)
@@ -164,7 +168,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.DomainBrowser
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # record browsing
 dbus (receive)
@@ -176,7 +180,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.RecordBrowser
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # service browsing
 dbus (receive)
@@ -188,7 +192,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.ServiceBrowser
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 
 # service type browsing
 dbus (receive)
@@ -200,7 +204,7 @@ dbus (receive)
 dbus (send)
     bus=system
     interface=org.freedesktop.Avahi.ServiceTypeBrowser
-    peer=(name=org.freedesktop.Avahi, label=###PLUG_SECURITY_TAGS###),
+    peer=(label=###PLUG_SECURITY_TAGS###),
 `
 
 // Note: avahiObservePermanentSlotDBus is used by avahi-control in DBusPermanentSlot
@@ -232,6 +236,8 @@ const avahiObserveConnectedPlugAppArmor = `
 # Description: allows domain, record, service, and service type browsing
 # as well as address, host and service resolving
 
+/{,var/}run/avahi-daemon/socket rw,
+
 #include <abstractions/dbus-strict>
 dbus (send)
     bus=system
@@ -256,12 +262,12 @@ dbus (receive)
 
 # Don't allow introspection since it reveals too much (path is not service
 # specific for unconfined)
+# do not use peer=(label=unconfined) here since this is DBus activated
 #dbus (send)
 #    bus=system
 #    path=/
 #    interface=org.freedesktop.DBus.Introspectable
-#    member=Introspect
-#    peer=(label=unconfined),
+#    member=Introspect,
 
 # These allows tampering with other snap's browsers, so don't autoconnect for
 # now.
@@ -421,7 +427,7 @@ func (iface *avahiObserveInterface) StaticInfo() interfaces.StaticInfo {
 	}
 }
 
-func (iface *avahiObserveInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
+func (iface *avahiObserveInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
 	old := "###SLOT_SECURITY_TAGS###"
 	var new string
 	if release.OnClassic {
@@ -443,7 +449,7 @@ func (iface *avahiObserveInterface) AppArmorPermanentSlot(spec *apparmor.Specifi
 	return nil
 }
 
-func (iface *avahiObserveInterface) AppArmorConnectedSlot(spec *apparmor.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
+func (iface *avahiObserveInterface) AppArmorConnectedSlot(spec *apparmor.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
 	if !release.OnClassic {
 		old := "###PLUG_SECURITY_TAGS###"
 		new := plugAppLabelExpr(plug)
@@ -460,7 +466,7 @@ func (iface *avahiObserveInterface) DBusPermanentSlot(spec *dbus.Specification,
 	return nil
 }
 
-func (iface *avahiObserveInterface) AutoConnect(*interfaces.Plug, *interfaces.Slot) bool {
+func (iface *avahiObserveInterface) AutoConnect(*snap.PlugInfo, *snap.SlotInfo) bool {
 	// allow what declarations allowed
 	return true
 }
