@@ -25,6 +25,7 @@ import (
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/apparmor"
 	"github.com/snapcore/snapd/interfaces/seccomp"
+	"github.com/snapcore/snapd/snap"
 )
 
 const browserSupportSummary = `allows access to various APIs needed by modern web browsers`
@@ -56,8 +57,8 @@ owner @{PROC}/@{pid}/oom_score_adj rw,
 /var/tmp/ r,
 owner /var/tmp/etilqs_* rw,
 
-# Chrome/Chromium should be modified to use snap.$SNAP_NAME.* or the snap
-# packaging adjusted to use LD_PRELOAD technique from LP: #1577514
+# Chrome/Chromium should be modified to use snap.$SNAP_INSTANCE_NAME.* or
+# the snap packaging adjusted to use LD_PRELOAD technique from LP: #1577514
 owner /{dev,run}/shm/{,.}org.chromium.* mrw,
 owner /{dev,run}/shm/{,.}com.google.Chrome.* mrw,
 owner /{dev,run}/shm/.io.nwjs.* mrw,
@@ -65,8 +66,9 @@ owner /{dev,run}/shm/.io.nwjs.* mrw,
 # Chrome's Singleton API sometimes causes an ouid/fsuid mismatch denial, so
 # for now, allow non-owner read on the singleton socket (LP: #1731012). See
 # https://forum.snapcraft.io/t/electron-snap-killed-when-using-app-makesingleinstance-api/2667/20
-/run/user/[0-9]*/snap.@{SNAP_NAME}/{,.}org.chromium.*/SS r,
-/run/user/[0-9]*/snap.@{SNAP_NAME}/{,.}com.google.Chrome.*/SS r,
+# parallel-installs: $XDG_RUNTIME_DIR is not remapped, need to use SNAP_INSTANCE_NAME
+/run/user/[0-9]*/snap.@{SNAP_INSTANCE_NAME}/{,.}org.chromium.*/SS r,
+/run/user/[0-9]*/snap.@{SNAP_INSTANCE_NAME}/{,.}com.google.Chrome.*/SS r,
 
 # Allow reading platform files
 /run/udev/data/+platform:* r,
@@ -104,19 +106,6 @@ owner @{PROC}/@{pid}/mountinfo r,
 deny capability mknod,
 `
 
-const browserSupportConnectedPlugAppArmorWithoutSandbox = `
-# ptrace can be used to break out of the seccomp sandbox, but ps requests
-# 'ptrace (trace)' even though it isn't tracing other processes. Unfortunately,
-# this is due to the kernel overloading trace such that the LSMs are unable to
-# distinguish between tracing other processes and other accesses. We deny the
-# trace here to silence the log.
-# Note: for now, explicitly deny to avoid confusion and accidentally giving
-# away this dangerous access frivolously. We may conditionally deny this in the
-# future. If the kernel has https://lkml.org/lkml/2016/5/26/354 we could also
-# allow this.
-deny ptrace (trace) peer=snap.@{SNAP_NAME}.**,
-`
-
 const browserSupportConnectedPlugAppArmorWithSandbox = `
 # Leaks installed applications
 # TODO: should this be somewhere else?
@@ -190,7 +179,7 @@ deny /sys/devices/virtual/block/dm-[0-9]*/dm/name r,
 /run/udev/data/b252:[0-9]* r,
 /run/udev/data/b253:[0-9]* r,
 /run/udev/data/b259:[0-9]* r,
-/run/udev/data/c24[2-9]:[0-9]* r,
+/run/udev/data/c24[0-9]:[0-9]* r,
 /run/udev/data/c25[0-4]:[0-9]* r,
 
 /sys/bus/**/devices/ r,
@@ -202,8 +191,8 @@ unix (bind)
 
 # Policy needed only when using the chrome/chromium setuid sandbox
 capability sys_ptrace,
-ptrace (trace) peer=snap.@{SNAP_NAME}.**,
-unix (receive, send) peer=(label=snap.@{SNAP_NAME}.**),
+ptrace (trace) peer=snap.@{SNAP_INSTANCE_NAME}.**,
+unix (receive, send) peer=(label=snap.@{SNAP_INSTANCE_NAME}.**),
 
 # If this were going to be allowed to all snaps, then for all the following
 # rules we would want to wrap in a 'browser_sandbox' profile, but a limitation
@@ -213,7 +202,8 @@ unix (receive, send) peer=(label=snap.@{SNAP_NAME}.**),
 # profile browser_sandbox {
 #   ...
 #   # This rule needs to work but generates a parser error
-#   @{INSTALL_DIR}/@{SNAP_NAME}/@{SNAP_REVISION}/opt/google/chrome/chrome px -> snap.@{SNAP_NAME}.@{SNAP_APP},
+#   @{INSTALL_DIR}/@{SNAP_NAME}/@{SNAP_REVISION}/opt/google/chrome/chrome px -> snap.@{SNAP_INSTANCE_NAME}.@{SNAP_APP},
+#   @{INSTALL_DIR}/@{SNAP_INSTANCE_NAME}/@{SNAP_REVISION}/opt/google/chrome/chrome px -> snap.@{SNAP_INSTANCE_NAME}.@{SNAP_APP},
 #   ...
 # }
 
@@ -282,6 +272,23 @@ setgid
 # Policy needed for Mozilla userns sandbox
 unshare
 quotactl
+
+# The Breakpad crash reporter uses ptrace to read register/memory state
+# from the crashed process, but it doesn't need to modify any state; see
+# https://bugzilla.mozilla.org/show_bug.cgi?id=1461848.
+#
+# These rules allow that but don't allow ptrace operations that
+# write registers, which can be used to bypass security; see
+# https://lkml.org/lkml/2016/5/26/354.
+ptrace PTRACE_ATTACH
+ptrace PTRACE_DETACH
+ptrace PTRACE_GETREGS
+ptrace PTRACE_GETFPREGS
+ptrace PTRACE_GETFPXREGS
+ptrace PTRACE_GETREGSET
+ptrace PTRACE_PEEKDATA
+ptrace PTRACE_PEEKUSER
+ptrace PTRACE_CONT
 `
 
 type browserSupportInterface struct{}
@@ -299,7 +306,7 @@ func (iface *browserSupportInterface) StaticInfo() interfaces.StaticInfo {
 	}
 }
 
-func (iface *browserSupportInterface) SanitizePlug(plug *interfaces.Plug) error {
+func (iface *browserSupportInterface) BeforePreparePlug(plug *snap.PlugInfo) error {
 	// It's fine if allow-sandbox isn't specified, but it it is,
 	// it needs to be bool
 	if v, ok := plug.Attrs["allow-sandbox"]; ok {
@@ -311,19 +318,21 @@ func (iface *browserSupportInterface) SanitizePlug(plug *interfaces.Plug) error
 	return nil
 }
 
-func (iface *browserSupportInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
-	allowSandbox, _ := plug.Attrs["allow-sandbox"].(bool)
+func (iface *browserSupportInterface) AppArmorConnectedPlug(spec *apparmor.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
+	var allowSandbox bool
+	_ = plug.Attr("allow-sandbox", &allowSandbox)
 	spec.AddSnippet(browserSupportConnectedPlugAppArmor)
 	if allowSandbox {
 		spec.AddSnippet(browserSupportConnectedPlugAppArmorWithSandbox)
 	} else {
-		spec.AddSnippet(browserSupportConnectedPlugAppArmorWithoutSandbox)
+		spec.SetSuppressPtraceTrace()
 	}
 	return nil
 }
 
-func (iface *browserSupportInterface) SecCompConnectedPlug(spec *seccomp.Specification, plug *interfaces.Plug, plugAttrs map[string]interface{}, slot *interfaces.Slot, slotAttrs map[string]interface{}) error {
-	allowSandbox, _ := plug.Attrs["allow-sandbox"].(bool)
+func (iface *browserSupportInterface) SecCompConnectedPlug(spec *seccomp.Specification, plug *interfaces.ConnectedPlug, slot *interfaces.ConnectedSlot) error {
+	var allowSandbox bool
+	_ = plug.Attr("allow-sandbox", &allowSandbox)
 	snippet := browserSupportConnectedPlugSecComp
 	if allowSandbox {
 		snippet += browserSupportConnectedPlugSecCompWithSandbox
@@ -332,7 +341,7 @@ func (iface *browserSupportInterface) SecCompConnectedPlug(spec *seccomp.Specifi
 	return nil
 }
 
-func (iface *browserSupportInterface) AutoConnect(*interfaces.Plug, *interfaces.Slot) bool {
+func (iface *browserSupportInterface) AutoConnect(*snap.PlugInfo, *snap.SlotInfo) bool {
 	return true
 }
 
