@@ -93,11 +93,11 @@ func (s *snapmgrTestSuite) TestApplyAliasesChange(c *C) {
 
 		var add, rm []*backend.Alias
 		if strings.Contains(scenario.ops, "rm") {
-			rm = []*backend.Alias{{"myalias", fmt.Sprintf("alias-snap1.%s", target(scenario.target))}}
+			rm = []*backend.Alias{{Name: "myalias", Target: fmt.Sprintf("alias-snap1.%s", target(scenario.target))}}
 		}
 
 		if strings.Contains(scenario.ops, "add") {
-			add = []*backend.Alias{{"myalias", fmt.Sprintf("alias-snap1.%s", target(scenario.newTarget))}}
+			add = []*backend.Alias{{Name: "myalias", Target: fmt.Sprintf("alias-snap1.%s", target(scenario.newTarget))}}
 		}
 
 		expected := fakeOps{
@@ -133,8 +133,8 @@ func (s *snapmgrTestSuite) TestApplyAliasesChangeMulti(c *C) {
 	expected := fakeOps{
 		{
 			op:        "update-aliases",
-			rmAliases: []*backend.Alias{{"myalias0", "alias-snap1.cmd0"}},
-			aliases:   []*backend.Alias{{"myalias1", "alias-snap1"}},
+			rmAliases: []*backend.Alias{{Name: "myalias0", Target: "alias-snap1.cmd0"}},
+			aliases:   []*backend.Alias{{Name: "myalias1", Target: "alias-snap1"}},
 		},
 	}
 
@@ -145,7 +145,7 @@ func (s *snapmgrTestSuite) TestApplyAliasesChangeMulti(c *C) {
 
 func (s *snapmgrTestSuite) TestAutoAliasesDelta(c *C) {
 	snapstate.AutoAliases = func(st *state.State, info *snap.Info) (map[string]string, error) {
-		c.Check(info.Name(), Equals, "alias-snap")
+		c.Check(info.InstanceName(), Equals, "alias-snap")
 		return map[string]string{
 			"alias1": "cmd1",
 			"alias2": "cmd2",
@@ -188,8 +188,8 @@ func (s *snapmgrTestSuite) TestAutoAliasesDelta(c *C) {
 func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 	seen := make(map[string]bool)
 	snapstate.AutoAliases = func(st *state.State, info *snap.Info) (map[string]string, error) {
-		seen[info.Name()] = true
-		if info.Name() == "alias-snap" {
+		seen[info.InstanceName()] = true
+		if info.InstanceName() == "alias-snap" {
 			return map[string]string{
 				"alias1": "cmd1",
 				"alias2": "cmd2",
@@ -229,6 +229,7 @@ func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 	c.Check(dropped, HasLen, 0)
 
 	c.Check(seen, DeepEquals, map[string]bool{
+		"core":       true,
 		"alias-snap": true,
 		"other-snap": true,
 	})
@@ -236,7 +237,7 @@ func (s *snapmgrTestSuite) TestAutoAliasesDeltaAll(c *C) {
 
 func (s *snapmgrTestSuite) TestAutoAliasesDeltaOverManual(c *C) {
 	snapstate.AutoAliases = func(st *state.State, info *snap.Info) (map[string]string, error) {
-		c.Check(info.Name(), Equals, "alias-snap")
+		c.Check(info.InstanceName(), Equals, "alias-snap")
 		return map[string]string{
 			"alias1": "cmd1",
 			"alias2": "cmd2",
@@ -273,7 +274,7 @@ func (s *snapmgrTestSuite) TestRefreshAliases(c *C) {
 	defer s.state.Unlock()
 
 	snapstate.AutoAliases = func(st *state.State, info *snap.Info) (map[string]string, error) {
-		c.Check(info.Name(), Equals, "alias-snap")
+		c.Check(info.InstanceName(), Equals, "alias-snap")
 		return map[string]string{
 			"alias1": "cmd1",
 			"alias2": "cmd2",
@@ -284,6 +285,7 @@ func (s *snapmgrTestSuite) TestRefreshAliases(c *C) {
 
 	info := snaptest.MockInfo(c, `
 name: alias-snap
+version: 0
 apps:
     cmd1:
     cmd2:
@@ -512,6 +514,14 @@ func (s *snapmgrTestSuite) TestAliasTasks(c *C) {
 		Current: snap.R(11),
 		Active:  true,
 	})
+	snapstate.Set(s.state, "some-snap_instance", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "some-snap", Revision: snap.R(11)},
+		},
+		Current:     snap.R(11),
+		Active:      true,
+		InstanceKey: "instance",
+	})
 
 	ts, err := snapstate.Alias(s.state, "some-snap", "cmd1", "alias1")
 	c.Assert(err, IsNil)
@@ -520,6 +530,22 @@ func (s *snapmgrTestSuite) TestAliasTasks(c *C) {
 	c.Assert(taskKinds(ts.Tasks()), DeepEquals, []string{
 		"alias",
 	})
+
+	ts, err = snapstate.Alias(s.state, "some-snap_instance", "cmd1", "alias1")
+	c.Assert(err, IsNil)
+
+	c.Assert(s.state.TaskCount(), Equals, 2)
+	c.Assert(taskKinds(ts.Tasks()), DeepEquals, []string{
+		"alias",
+	})
+
+	var snapsup snapstate.SnapSetup
+	tasks := ts.Tasks()
+	c.Assert(tasks, HasLen, 1)
+	err = tasks[0].Get("snap-setup", &snapsup)
+	c.Assert(err, IsNil)
+	c.Check(snapsup.InstanceKey, Equals, "instance")
+	c.Check(snapsup.InstanceName(), Equals, "some-snap_instance")
 }
 
 type changedAlias struct {
@@ -551,15 +577,15 @@ func (s *snapmgrTestSuite) TestAliasRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
 		{
 			op:      "update-aliases",
-			aliases: []*backend.Alias{{"alias1", "alias-snap.cmd1"}},
+			aliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd1"}},
 		},
 	}
 	// start with an easier-to-read error if this fails:
@@ -584,6 +610,70 @@ func (s *snapmgrTestSuite) TestAliasRunThrough(c *C) {
 	})
 }
 
+func (s *snapmgrTestSuite) TestParallelInstanceAliasRunThrough(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "alias-snap", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+	})
+	snapstate.Set(s.state, "alias-snap_foo", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current:     snap.R(11),
+		Active:      true,
+		InstanceKey: "foo",
+	})
+
+	chg := s.state.NewChange("alias", "manual alias")
+	ts, err := snapstate.Alias(s.state, "alias-snap_foo", "cmd1", "alias1")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	defer s.se.Stop()
+	s.settle(c)
+	s.state.Lock()
+
+	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
+	expected := fakeOps{
+		{
+			op:      "update-aliases",
+			aliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap_foo.cmd1"}},
+		},
+	}
+	// start with an easier-to-read error if this fails:
+	c.Assert(s.fakeBackend.ops.Ops(), DeepEquals, expected.Ops())
+	c.Assert(s.fakeBackend.ops, DeepEquals, expected)
+
+	var snapst snapstate.SnapState
+	err = snapstate.Get(s.state, "alias-snap_foo", &snapst)
+	c.Assert(err, IsNil)
+
+	c.Check(snapst.AutoAliasesDisabled, Equals, false)
+	c.Check(snapst.AliasesPending, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Manual: "cmd1"},
+	})
+
+	// snap without instance key is unchanged
+	err = snapstate.Get(s.state, "alias-snap", &snapst)
+	c.Assert(err, IsNil)
+	c.Check(snapst.Aliases, HasLen, 0)
+
+	var trace traceData
+	err = chg.Get("api-data", &trace)
+	c.Assert(err, IsNil)
+	c.Check(trace, DeepEquals, traceData{
+		Added: []*changedAlias{{Snap: "alias-snap_foo", App: "cmd1", Alias: "alias1"}},
+	})
+}
+
 func (s *snapmgrTestSuite) TestAliasNoTarget(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -602,8 +692,8 @@ func (s *snapmgrTestSuite) TestAliasNoTarget(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	s.snapmgr.Ensure()
-	s.snapmgr.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 	s.state.Lock()
 
 	c.Check(chg.Status(), Equals, state.ErrorStatus)
@@ -628,8 +718,8 @@ func (s *snapmgrTestSuite) TestAliasTargetIsDaemon(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	s.snapmgr.Ensure()
-	s.snapmgr.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 	s.state.Lock()
 
 	c.Check(chg.Status(), Equals, state.ErrorStatus)
@@ -666,16 +756,16 @@ func (s *snapmgrTestSuite) TestAliasOverAutoRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
 		{
 			op:        "update-aliases",
-			rmAliases: []*backend.Alias{{"alias1", "alias-snap.cmd1"}},
-			aliases:   []*backend.Alias{{"alias1", "alias-snap.cmd5"}},
+			rmAliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd1"}},
+			aliases:   []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd5"}},
 		},
 	}
 	// start with an easier-to-read error if this fails:
@@ -711,7 +801,7 @@ func (s *snapmgrTestSuite) TestAliasUpdateChangeConflict(c *C) {
 	s.state.NewChange("alias", "...").AddAll(ts)
 
 	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "alias" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestUpdateAliasChangeConflict(c *C) {
@@ -731,7 +821,7 @@ func (s *snapmgrTestSuite) TestUpdateAliasChangeConflict(c *C) {
 	s.state.NewChange("update", "...").AddAll(ts)
 
 	_, err = snapstate.Alias(s.state, "some-snap", "cmd1", "alias1")
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "update" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestAliasAliasConflict(c *C) {
@@ -762,14 +852,69 @@ func (s *snapmgrTestSuite) TestAliasAliasConflict(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	s.snapmgr.Ensure()
-	s.snapmgr.Wait()
+	s.se.Ensure()
+	s.se.Wait()
 	s.state.Lock()
 
 	c.Check(chg.Status(), Equals, state.ErrorStatus)
 	c.Check(chg.Err(), ErrorMatches, `(?s).*cannot enable alias "alias1" for "alias-snap", already enabled for "other-snap".*`)
 }
 
+func (s *snapmgrTestSuite) TestParallelInstanceAliasConflict(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "alias-snap", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias2": {Auto: "cmd2"},
+		},
+	})
+	snapstate.Set(s.state, "alias-snap_foo", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current:     snap.R(11),
+		Active:      true,
+		InstanceKey: "foo",
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Auto: "cmd1"},
+		},
+	})
+
+	chg := s.state.NewChange("alias", "alias")
+	// conflicts with aliases of alias-snap_foo
+	ts, err := snapstate.Alias(s.state, "alias-snap", "cmd5", "alias1")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	s.se.Ensure()
+	s.se.Wait()
+	s.state.Lock()
+
+	c.Check(chg.Status(), Equals, state.ErrorStatus)
+	c.Check(chg.Err(), ErrorMatches, `(?s).*cannot enable alias "alias1" for "alias-snap", already enabled for "alias-snap_foo".*`)
+
+	chg = s.state.NewChange("alias", "alias")
+	// conflicts with aliases of alias-snap
+	ts, err = snapstate.Alias(s.state, "alias-snap_foo", "cmd5", "alias2")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	s.se.Ensure()
+	s.se.Wait()
+	s.state.Lock()
+
+	c.Check(chg.Status(), Equals, state.ErrorStatus)
+	c.Check(chg.Err(), ErrorMatches, `(?s).*cannot enable alias "alias2" for "alias-snap_foo", already enabled for "alias-snap".*`)
+}
+
 func (s *snapmgrTestSuite) TestDisableAllAliasesTasks(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -814,8 +959,8 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
@@ -823,9 +968,9 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesRunThrough(c *C) {
 		{
 			op: "update-aliases",
 			rmAliases: []*backend.Alias{
-				{"alias1", "alias-snap.cmd5"},
-				{"alias2", "alias-snap.cmd2"},
-				{"alias3", "alias-snap.cmd3"},
+				{Name: "alias1", Target: "alias-snap.cmd5"},
+				{Name: "alias2", Target: "alias-snap.cmd2"},
+				{Name: "alias3", Target: "alias-snap.cmd3"},
 			},
 		},
 	}
@@ -851,6 +996,87 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesRunThrough(c *C) {
 	c.Check(trace.Removed, HasLen, 3)
 }
 
+func (s *snapmgrTestSuite) TestParallelInstanceDisableAllAliasesRunThrough(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "alias-snap_instance", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Auto: "cmd1"},
+			"alias2": {Auto: "cmd2"},
+			"alias3": {Manual: "cmd3"},
+		},
+		InstanceKey: "instance",
+	})
+	snapstate.Set(s.state, "alias-snap", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Manual: "cmd6"},
+			"alias2": {Manual: "cmd7"},
+		},
+	})
+
+	chg := s.state.NewChange("unalias", "unalias")
+	ts, err := snapstate.DisableAllAliases(s.state, "alias-snap_instance")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	defer s.se.Stop()
+	s.settle(c)
+	s.state.Lock()
+
+	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
+	expected := fakeOps{
+		{
+			op: "update-aliases",
+			rmAliases: []*backend.Alias{
+				{Name: "alias1", Target: "alias-snap_instance.cmd1"},
+				{Name: "alias2", Target: "alias-snap_instance.cmd2"},
+				{Name: "alias3", Target: "alias-snap_instance.cmd3"},
+			},
+		},
+	}
+	// start with an easier-to-read error if this fails:
+	c.Assert(s.fakeBackend.ops.Ops(), DeepEquals, expected.Ops())
+	c.Assert(s.fakeBackend.ops, DeepEquals, expected)
+
+	var snapst snapstate.SnapState
+	err = snapstate.Get(s.state, "alias-snap_instance", &snapst)
+	c.Assert(err, IsNil)
+
+	c.Check(snapst.AutoAliasesDisabled, Equals, true)
+	c.Check(snapst.AliasesPending, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Auto: "cmd1"},
+		"alias2": {Auto: "cmd2"},
+	})
+
+	var trace traceData
+	err = chg.Get("api-data", &trace)
+	c.Assert(err, IsNil)
+	c.Check(trace.Added, HasLen, 0)
+	c.Check(trace.Removed, HasLen, 3)
+
+	// non _instance instance is unchanged
+	err = snapstate.Get(s.state, "alias-snap", &snapst)
+	c.Assert(err, IsNil)
+	c.Check(snapst.AutoAliasesDisabled, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Manual: "cmd6"},
+		"alias2": {Manual: "cmd7"},
+	})
+}
+
 func (s *snapmgrTestSuite) TestRemoveManualAliasTasks(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -902,15 +1128,15 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
 		{
 			op:        "update-aliases",
-			rmAliases: []*backend.Alias{{"alias1", "alias-snap.cmd5"}},
+			rmAliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd5"}},
 		},
 	}
 	// start with an easier-to-read error if this fails:
@@ -954,16 +1180,16 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasOverAutoRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
 	expected := fakeOps{
 		{
 			op:        "update-aliases",
-			rmAliases: []*backend.Alias{{"alias1", "alias-snap.cmd5"}},
-			aliases:   []*backend.Alias{{"alias1", "alias-snap.cmd1"}},
+			rmAliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd5"}},
+			aliases:   []*backend.Alias{{Name: "alias1", Target: "alias-snap.cmd1"}},
 		},
 	}
 	// start with an easier-to-read error if this fails:
@@ -981,6 +1207,55 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasOverAutoRunThrough(c *C) {
 	})
 }
 
+func (s *snapmgrTestSuite) TestParallelInstanceRemoveManualAliasRunThrough(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "alias-snap_instance", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Manual: "cmd2", Auto: "cmd1"},
+		},
+		InstanceKey: "instance",
+	})
+
+	chg := s.state.NewChange("alias", "manual alias")
+	ts, _, err := snapstate.RemoveManualAlias(s.state, "alias1")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	defer s.se.Stop()
+	s.settle(c)
+	s.state.Lock()
+
+	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
+	expected := fakeOps{
+		{
+			op:        "update-aliases",
+			rmAliases: []*backend.Alias{{Name: "alias1", Target: "alias-snap_instance.cmd2"}},
+			aliases:   []*backend.Alias{{Name: "alias1", Target: "alias-snap_instance.cmd1"}},
+		},
+	}
+	// start with an easier-to-read error if this fails:
+	c.Assert(s.fakeBackend.ops.Ops(), DeepEquals, expected.Ops())
+	c.Assert(s.fakeBackend.ops, DeepEquals, expected)
+
+	var snapst snapstate.SnapState
+	err = snapstate.Get(s.state, "alias-snap_instance", &snapst)
+	c.Assert(err, IsNil)
+
+	c.Check(snapst.AutoAliasesDisabled, Equals, false)
+	c.Check(snapst.AliasesPending, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Auto: "cmd1"},
+	})
+}
+
 func (s *snapmgrTestSuite) TestRemoveManualAliasNoAlias(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -1009,7 +1284,7 @@ func (s *snapmgrTestSuite) TestUpdateDisableAllAliasesChangeConflict(c *C) {
 	s.state.NewChange("update", "...").AddAll(ts)
 
 	_, err = snapstate.DisableAllAliases(s.state, "some-snap")
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "update" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestUpdateRemoveManualAliasChangeConflict(c *C) {
@@ -1032,7 +1307,7 @@ func (s *snapmgrTestSuite) TestUpdateRemoveManualAliasChangeConflict(c *C) {
 	s.state.NewChange("update", "...").AddAll(ts)
 
 	_, _, err = snapstate.RemoveManualAlias(s.state, "alias1")
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "update" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestDisableAllAliasesUpdateChangeConflict(c *C) {
@@ -1052,7 +1327,7 @@ func (s *snapmgrTestSuite) TestDisableAllAliasesUpdateChangeConflict(c *C) {
 	s.state.NewChange("alias", "...").AddAll(ts)
 
 	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "alias" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestRemoveManualAliasUpdateChangeConflict(c *C) {
@@ -1075,7 +1350,7 @@ func (s *snapmgrTestSuite) TestRemoveManualAliasUpdateChangeConflict(c *C) {
 	s.state.NewChange("unalias", "...").AddAll(ts)
 
 	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "unalias" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestPreferTasks(c *C) {
@@ -1122,8 +1397,8 @@ func (s *snapmgrTestSuite) TestPreferRunThrough(c *C) {
 	chg.AddAll(ts)
 
 	s.state.Unlock()
-	defer s.snapmgr.Stop()
-	s.settle()
+	defer s.se.Stop()
+	s.settle(c)
 	s.state.Lock()
 
 	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
@@ -1131,8 +1406,8 @@ func (s *snapmgrTestSuite) TestPreferRunThrough(c *C) {
 		{
 			op: "update-aliases",
 			aliases: []*backend.Alias{
-				{"alias1", "alias-snap.cmd1"},
-				{"alias2", "alias-snap.cmd2"},
+				{Name: "alias1", Target: "alias-snap.cmd1"},
+				{Name: "alias2", Target: "alias-snap.cmd2"},
 			},
 		},
 	}
@@ -1152,6 +1427,87 @@ func (s *snapmgrTestSuite) TestPreferRunThrough(c *C) {
 	})
 }
 
+func (s *snapmgrTestSuite) TestParallelInstancePreferRunThrough(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapstate.Set(s.state, "alias-snap", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current: snap.R(11),
+		Active:  true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Auto: "cmd1"},
+			"alias2": {Auto: "cmd2"},
+		},
+	})
+	snapstate.Set(s.state, "alias-snap_instance", &snapstate.SnapState{
+		Sequence: []*snap.SideInfo{
+			{RealName: "alias-snap", Revision: snap.R(11)},
+		},
+		Current:             snap.R(11),
+		Active:              true,
+		AutoAliasesDisabled: true,
+		Aliases: map[string]*snapstate.AliasTarget{
+			"alias1": {Auto: "cmd1"},
+			"alias2": {Auto: "cmd2"},
+		},
+	})
+
+	chg := s.state.NewChange("prefer", "prefer")
+	ts, err := snapstate.Prefer(s.state, "alias-snap_instance")
+	c.Assert(err, IsNil)
+	chg.AddAll(ts)
+
+	s.state.Unlock()
+	defer s.se.Stop()
+	s.settle(c)
+	s.state.Lock()
+
+	c.Assert(chg.Status(), Equals, state.DoneStatus, Commentf("%v", chg.Err()))
+	expected := fakeOps{
+		{
+			op: "update-aliases",
+			rmAliases: []*backend.Alias{
+				{Name: "alias1", Target: "alias-snap.cmd1"},
+				{Name: "alias2", Target: "alias-snap.cmd2"},
+			},
+		},
+		{
+			op: "update-aliases",
+			aliases: []*backend.Alias{
+				{Name: "alias1", Target: "alias-snap_instance.cmd1"},
+				{Name: "alias2", Target: "alias-snap_instance.cmd2"},
+			},
+		},
+	}
+	// start with an easier-to-read error if this fails:
+	c.Assert(s.fakeBackend.ops.Ops(), DeepEquals, expected.Ops())
+	c.Assert(s.fakeBackend.ops, DeepEquals, expected)
+
+	var snapst snapstate.SnapState
+	err = snapstate.Get(s.state, "alias-snap_instance", &snapst)
+	c.Assert(err, IsNil)
+	// _instance aliases are preferred now
+	c.Check(snapst.AutoAliasesDisabled, Equals, false)
+	c.Check(snapst.AliasesPending, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Auto: "cmd1"},
+		"alias2": {Auto: "cmd2"},
+	})
+
+	err = snapstate.Get(s.state, "alias-snap", &snapst)
+	c.Assert(err, IsNil)
+	c.Check(snapst.AutoAliasesDisabled, Equals, true)
+	c.Check(snapst.AliasesPending, Equals, false)
+	c.Check(snapst.Aliases, DeepEquals, map[string]*snapstate.AliasTarget{
+		"alias1": {Auto: "cmd1"},
+		"alias2": {Auto: "cmd2"},
+	})
+
+}
+
 func (s *snapmgrTestSuite) TestUpdatePreferChangeConflict(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -1173,7 +1529,7 @@ func (s *snapmgrTestSuite) TestUpdatePreferChangeConflict(c *C) {
 	s.state.NewChange("update", "...").AddAll(ts)
 
 	_, err = snapstate.Prefer(s.state, "some-snap")
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "update" change in progress`)
 }
 
 func (s *snapmgrTestSuite) TestPreferUpdateChangeConflict(c *C) {
@@ -1197,5 +1553,5 @@ func (s *snapmgrTestSuite) TestPreferUpdateChangeConflict(c *C) {
 	s.state.NewChange("prefer", "...").AddAll(ts)
 
 	_, err = snapstate.Update(s.state, "some-snap", "some-channel", snap.R(0), s.user.ID, snapstate.Flags{})
-	c.Assert(err, ErrorMatches, `snap "some-snap" has changes in progress`)
+	c.Assert(err, ErrorMatches, `snap "some-snap" has "prefer" change in progress`)
 }
