@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2017 Canonical Ltd
+ * Copyright (C) 2016-2018 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,8 +20,10 @@
 package snapstate
 
 import (
+	"encoding/json"
 	"fmt"
 	"os"
+	"path/filepath"
 	"strconv"
 	"strings"
 	"time"
@@ -29,13 +31,19 @@ import (
 	"gopkg.in/tomb.v2"
 
 	"github.com/snapcore/snapd/boot"
+	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/logger"
+	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/configstate/config"
+	"github.com/snapcore/snapd/overlord/configstate/settings"
 	"github.com/snapcore/snapd/overlord/snapstate/backend"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
+	"github.com/snapcore/snapd/strutil"
 )
 
 // TaskSnapSetup returns the SnapSetup with task params hold by or referred to by the the task.
@@ -57,6 +65,9 @@ func TaskSnapSetup(t *state.Task) (*SnapSetup, error) {
 	}
 
 	ts := t.State().Task(id)
+	if ts == nil {
+		return nil, fmt.Errorf("internal error: tasks are being pruned")
+	}
 	if err := ts.Get("snap-setup", &snapsup); err != nil {
 		return nil, err
 	}
@@ -69,7 +80,7 @@ func snapSetupAndState(t *state.Task) (*SnapSetup, *SnapState, error) {
 		return nil, nil, err
 	}
 	var snapst SnapState
-	err = Get(t.State(), snapsup.Name(), &snapst)
+	err = Get(t.State(), snapsup.InstanceName(), &snapst)
 	if err != nil && err != state.ErrNoState {
 		return nil, nil, err
 	}
@@ -112,21 +123,38 @@ func snapSetupAndState(t *state.Task) (*SnapSetup, *SnapState, error) {
 */
 
 const defaultCoreSnapName = "core"
-const defaultBaseSnapsChannel = "stable"
 
-func changeInFlight(st *state.State, snapName string) (bool, error) {
+func defaultBaseSnapsChannel() string {
+	channel := os.Getenv("SNAPD_BASES_CHANNEL")
+	if channel == "" {
+		return "stable"
+	}
+	return channel
+}
+
+func defaultPrereqSnapsChannel() string {
+	channel := os.Getenv("SNAPD_PREREQS_CHANNEL")
+	if channel == "" {
+		return "stable"
+	}
+	return channel
+}
+
+func linkSnapInFlight(st *state.State, snapName string) (bool, error) {
 	for _, chg := range st.Changes() {
 		if chg.Status().Ready() {
 			continue
 		}
 		for _, tc := range chg.Tasks() {
-			if tc.Kind() == "link-snap" || tc.Kind() == "discard-snap" {
+			if tc.Status().Ready() {
+				continue
+			}
+			if tc.Kind() == "link-snap" {
 				snapsup, err := TaskSnapSetup(tc)
 				if err != nil {
 					return false, err
 				}
-				// some other change aleady inflight
-				if snapsup.Name() == snapName {
+				if snapsup.InstanceName() == snapName {
 					return true, nil
 				}
 			}
@@ -136,6 +164,15 @@ func changeInFlight(st *state.State, snapName string) (bool, error) {
 	return false, nil
 }
 
+func isInstalled(st *state.State, snapName string) (bool, error) {
+	var snapState SnapState
+	err := Get(st, snapName, &snapState)
+	if err != nil && err != state.ErrNoState {
+		return false, err
+	}
+	return snapState.IsInstalled(), nil
+}
+
 // timeout for tasks to check if the prerequisites are ready
 var prerequisitesRetryTimeout = 30 * time.Second
 
@@ -150,61 +187,103 @@ func (m *SnapManager) doPrerequisites(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	// core/ubuntu-core can not have prerequisites
-	snapName := snapsup.Name()
-	if snapName == defaultCoreSnapName || snapName == "ubuntu-core" {
+	// os/base/kernel/gadget cannot have prerequisites other
+	// than the models default base (or core) which is installed anyway
+	switch snapsup.Type {
+	case snap.TypeOS, snap.TypeBase, snap.TypeKernel, snap.TypeGadget:
+		return nil
+	}
+	// snapd is special and has no prereqs
+	if snapsup.InstanceName() == "snapd" {
 		return nil
 	}
 
-	// check prereqs
-	prereqName := defaultCoreSnapName
+	// we need to make sure we install all prereqs together in one
+	// operation
+	base := defaultCoreSnapName
 	if snapsup.Base != "" {
-		prereqName = snapsup.Base
+		base = snapsup.Base
 	}
 
-	var prereqState SnapState
-	err = Get(st, prereqName, &prereqState)
-	// we have the prereq already
-	if err == nil {
-		return nil
-	}
-	// if it is a real error, report
-	if err != state.ErrNoState {
+	if err := m.installPrereqs(t, base, snapsup.Prereq, snapsup.UserID); err != nil {
 		return err
 	}
 
-	// check that there is no task that installs the prereq already
-	prereqPending, err := changeInFlight(st, prereqName)
+	return nil
+}
+
+func (m *SnapManager) installOneBaseOrRequired(st *state.State, snapName, channel string, onInFlight error, userID int) (*state.TaskSet, error) {
+	// installed already?
+	isInstalled, err := isInstalled(st, snapName)
 	if err != nil {
-		return err
+		return nil, err
 	}
-	if prereqPending {
-		// if something else installs core already we need to
-		// wait for that to either finish successfully or fail
-		return &state.Retry{After: prerequisitesRetryTimeout}
+	if isInstalled {
+		return nil, nil
+	}
+	// in progress?
+	inFlight, err := linkSnapInFlight(st, snapName)
+	if err != nil {
+		return nil, err
+	}
+	if inFlight {
+		return nil, onInFlight
 	}
 
 	// not installed, nor queued for install -> install it
-	ts, err := Install(st, prereqName, defaultBaseSnapsChannel, snap.R(0), snapsup.UserID, Flags{})
-	// something might have triggered an explicit install of core while
-	// the state was unlocked -> deal with that here
-	if _, ok := err.(changeDuringInstallError); ok {
-		return &state.Retry{After: prerequisitesRetryTimeout}
+	ts, err := Install(st, snapName, channel, snap.R(0), userID, Flags{})
+	// something might have triggered an explicit install while
+	// the state was unlocked -> deal with that here by simply
+	// retrying the operation.
+	if _, ok := err.(*ChangeConflictError); ok {
+		return nil, &state.Retry{After: prerequisitesRetryTimeout}
 	}
-	if _, ok := err.(changeConflictError); ok {
-		return &state.Retry{After: prerequisitesRetryTimeout}
+	return ts, err
+}
+
+func (m *SnapManager) installPrereqs(t *state.Task, base string, prereq []string, userID int) error {
+	st := t.State()
+
+	// We try to install all wanted snaps. If one snap cannot be installed
+	// because of change conflicts or similar we retry. Only if all snaps
+	// can be installed together we add the tasks to the change.
+	var tss []*state.TaskSet
+	for _, prereqName := range prereq {
+		var onInFlightErr error = nil
+		ts, err := m.installOneBaseOrRequired(st, prereqName, defaultPrereqSnapsChannel(), onInFlightErr, userID)
+		if err != nil {
+			return err
+		}
+		if ts == nil {
+			continue
+		}
+		tss = append(tss, ts)
 	}
+
+	// for base snaps we need to wait until the change is done
+	// (either finished or failed)
+	onInFlightErr := &state.Retry{After: prerequisitesRetryTimeout}
+	tsBase, err := m.installOneBaseOrRequired(st, base, defaultBaseSnapsChannel(), onInFlightErr, userID)
 	if err != nil {
 		return err
 	}
-	ts.JoinLane(st.NewLane())
 
-	// inject install for core into this change
 	chg := t.Change()
-	for _, t := range chg.Tasks() {
-		t.WaitAll(ts)
+	// add all required snaps, no ordering, this will be done in the
+	// auto-connect task handler
+	for _, ts := range tss {
+		ts.JoinLane(st.NewLane())
+		chg.AddAll(ts)
 	}
-	chg.AddAll(ts)
+	// add the base if needed, everything else must wait on this
+	if tsBase != nil {
+		tsBase.JoinLane(st.NewLane())
+		for _, t := range chg.Tasks() {
+			t.WaitAll(tsBase)
+		}
+		chg.AddAll(tsBase)
+	}
+
 	// make sure that the new change is committed to the state
 	// together with marking this task done
 	t.SetStatus(state.DoneStatus)
@@ -298,18 +377,52 @@ func (m *SnapManager) undoPrepareSnap(t *state.Task, _ *tomb.Tomb) error {
 		extra["UbuntuCoreTransitionCount"] = strconv.Itoa(ubuntuCoreTransitionCount)
 	}
 
-	st.Unlock()
-	oopsid, err := errtrackerReport(snapsup.SideInfo.RealName, strings.Join(logMsg, "\n"), strings.Join(dupSig, "\n"), extra)
-	st.Lock()
-	if err == nil {
-		logger.Noticef("Reported install problem for %q as %s", snapsup.SideInfo.RealName, oopsid)
-	} else {
-		logger.Debugf("Cannot report problem: %s", err)
+	// Only report and error if there is an actual error in the change,
+	// we could undo things because the user canceled the change.
+	var isErr bool
+	for _, tt := range t.Change().Tasks() {
+		if tt.Status() == state.ErrorStatus {
+			isErr = true
+			break
+		}
+	}
+	if isErr && !settings.ProblemReportsDisabled(st) {
+		st.Unlock()
+		oopsid, err := errtrackerReport(snapsup.SideInfo.RealName, strings.Join(logMsg, "\n"), strings.Join(dupSig, "\n"), extra)
+		st.Lock()
+		if err == nil {
+			logger.Noticef("Reported install problem for %q as %s", snapsup.SideInfo.RealName, oopsid)
+		} else {
+			logger.Debugf("Cannot report problem: %s", err)
+		}
 	}
 
 	return nil
 }
 
+func installInfoUnlocked(st *state.State, snapsup *SnapSetup) (*snap.Info, error) {
+	st.Lock()
+	defer st.Unlock()
+	return installInfo(st, snapsup.InstanceName(), snapsup.Channel, snapsup.Revision(), snapsup.UserID)
+}
+
+// autoRefreshRateLimited returns the rate limit of auto-refreshes or 0 if
+// there is no limit.
+func autoRefreshRateLimited(st *state.State) (rate int64) {
+	tr := config.NewTransaction(st)
+
+	var rateLimit string
+	err := tr.Get("core", "refresh.rate-limit", &rateLimit)
+	if err != nil {
+		return 0
+	}
+	val, err := strutil.ParseByteSize(rateLimit)
+	if err != nil {
+		return 0
+	}
+	return val
+}
+
 func (m *SnapManager) doDownloadSnap(t *state.Task, tomb *tomb.Tomb) error {
 	st := t.State()
 	st.Lock()
@@ -321,6 +434,7 @@ func (m *SnapManager) doDownloadSnap(t *state.Task, tomb *tomb.Tomb) error {
 
 	st.Lock()
 	theStore := Store(st)
+	rate := autoRefreshRateLimited(st)
 	user, err := userFromUserID(st, snapsup.UserID)
 	st.Unlock()
 	if err != nil {
@@ -329,24 +443,24 @@ func (m *SnapManager) doDownloadSnap(t *state.Task, tomb *tomb.Tomb) error {
 
 	meter := NewTaskProgressAdapterUnlocked(t)
 	targetFn := snapsup.MountFile()
+
+	dlOpts := &store.DownloadOptions{}
+	if snapsup.IsAutoRefresh && rate > 0 {
+		dlOpts.RateLimit = rate
+	}
 	if snapsup.DownloadInfo == nil {
 		var storeInfo *snap.Info
 		// COMPATIBILITY - this task was created from an older version
 		// of snapd that did not store the DownloadInfo in the state
 		// yet.
-		spec := store.SnapSpec{
-			Name:     snapsup.Name(),
-			Channel:  snapsup.Channel,
-			Revision: snapsup.Revision(),
-		}
-		storeInfo, err = theStore.SnapInfo(spec, user)
+		storeInfo, err = installInfoUnlocked(st, snapsup)
 		if err != nil {
 			return err
 		}
-		err = theStore.Download(tomb.Context(nil), snapsup.Name(), targetFn, &storeInfo.DownloadInfo, meter, user)
+		err = theStore.Download(tomb.Context(nil), snapsup.SnapName(), targetFn, &storeInfo.DownloadInfo, meter, user, dlOpts)
 		snapsup.SideInfo = &storeInfo.SideInfo
 	} else {
-		err = theStore.Download(tomb.Context(nil), snapsup.Name(), targetFn, snapsup.DownloadInfo, meter, user)
+		err = theStore.Download(tomb.Context(nil), snapsup.SnapName(), targetFn, snapsup.DownloadInfo, meter, user, dlOpts)
 	}
 	if err != nil {
 		return err
@@ -362,10 +476,34 @@ func (m *SnapManager) doDownloadSnap(t *state.Task, tomb *tomb.Tomb) error {
 	return nil
 }
 
+var (
+	mountPollInterval = 1 * time.Second
+)
+
+// hasOtherInstances checks whether there are other instances of the snap, be it
+// instance keyed or not
+func hasOtherInstances(st *state.State, instanceName string) (bool, error) {
+	snapName, _ := snap.SplitInstanceName(instanceName)
+	var all map[string]*json.RawMessage
+	if err := st.Get("snaps", &all); err != nil && err != state.ErrNoState {
+		return false, err
+	}
+	for otherName := range all {
+		if otherName == instanceName {
+			continue
+		}
+		if otherSnapName, _ := snap.SplitInstanceName(otherName); otherSnapName == snapName {
+			return true, nil
+		}
+	}
+	return false, nil
+}
+
 func (m *SnapManager) doMountSnap(t *state.Task, _ *tomb.Tomb) error {
-	t.State().Lock()
+	st := t.State()
+	st.Lock()
 	snapsup, snapst, err := snapSetupAndState(t)
-	t.State().Unlock()
+	st.Unlock()
 	if err != nil {
 		return err
 	}
@@ -376,26 +514,69 @@ func (m *SnapManager) doMountSnap(t *state.Task, _ *tomb.Tomb) error {
 
 	m.backend.CurrentInfo(curInfo)
 
-	if err := checkSnap(t.State(), snapsup.SnapPath, snapsup.SideInfo, curInfo, snapsup.Flags); err != nil {
+	if err := checkSnap(st, snapsup.SnapPath, snapsup.InstanceName(), snapsup.SideInfo, curInfo, snapsup.Flags); err != nil {
 		return err
 	}
 
+	cleanup := func() {
+		st.Lock()
+		defer st.Unlock()
+
+		otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+		if err != nil {
+			t.Errorf("cannot cleanup partial setup snap %q: %v", snapsup.InstanceName(), err)
+			return
+		}
+
+		// remove snap dir is idempotent so it's ok to always call it in the cleanup path
+		if err := m.backend.RemoveSnapDir(snapsup.placeInfo(), otherInstances); err != nil {
+			t.Errorf("cannot cleanup partial setup snap %q: %v", snapsup.InstanceName(), err)
+		}
+
+	}
+
 	pb := NewTaskProgressAdapterUnlocked(t)
 	// TODO Use snapsup.Revision() to obtain the right info to mount
 	//      instead of assuming the candidate is the right one.
-	if err := m.backend.SetupSnap(snapsup.SnapPath, snapsup.SideInfo, pb); err != nil {
-		return err
-	}
-
-	// set snapst type for undoMountSnap
-	newInfo, err := readInfo(snapsup.Name(), snapsup.SideInfo)
+	snapType, err := m.backend.SetupSnap(snapsup.SnapPath, snapsup.InstanceName(), snapsup.SideInfo, pb)
 	if err != nil {
+		cleanup()
 		return err
 	}
 
-	t.State().Lock()
-	t.Set("snap-type", newInfo.Type)
-	t.State().Unlock()
+	// double check that the snap is mounted
+	var readInfoErr error
+	for i := 0; i < 10; i++ {
+		_, readInfoErr = readInfo(snapsup.InstanceName(), snapsup.SideInfo, errorOnBroken)
+		if readInfoErr == nil {
+			break
+		}
+		if _, ok := readInfoErr.(*snap.NotFoundError); !ok {
+			break
+		}
+		// snap not found, seems is not mounted yet
+		msg := fmt.Sprintf("expected snap %q revision %v to be mounted but is not", snapsup.InstanceName(), snapsup.Revision())
+		readInfoErr = fmt.Errorf("cannot proceed, %s", msg)
+		if i == 0 {
+			logger.Noticef(msg)
+		}
+		time.Sleep(mountPollInterval)
+	}
+	if readInfoErr != nil {
+		if err := m.backend.UndoSetupSnap(snapsup.placeInfo(), snapType, pb); err != nil {
+			st.Lock()
+			t.Errorf("cannot undo partial setup snap %q: %v", snapsup.InstanceName(), err)
+			st.Unlock()
+		}
+
+		cleanup()
+		return readInfoErr
+	}
+
+	st.Lock()
+	// set snapst type for undoMountSnap
+	t.Set("snap-type", snapType)
+	st.Unlock()
 
 	if snapsup.Flags.RemoveSnapPath {
 		if err := os.Remove(snapsup.SnapPath); err != nil {
@@ -407,17 +588,18 @@ func (m *SnapManager) doMountSnap(t *state.Task, _ *tomb.Tomb) error {
 }
 
 func (m *SnapManager) undoMountSnap(t *state.Task, _ *tomb.Tomb) error {
-	t.State().Lock()
+	st := t.State()
+	st.Lock()
 	snapsup, err := TaskSnapSetup(t)
-	t.State().Unlock()
+	st.Unlock()
 	if err != nil {
 		return err
 	}
 
-	t.State().Lock()
+	st.Lock()
 	var typ snap.Type
 	err = t.Get("snap-type", &typ)
-	t.State().Unlock()
+	st.Unlock()
 	// backward compatibility
 	if err == state.ErrNoState {
 		typ = "app"
@@ -426,7 +608,19 @@ func (m *SnapManager) undoMountSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	pb := NewTaskProgressAdapterUnlocked(t)
-	return m.backend.UndoSetupSnap(snapsup.placeInfo(), typ, pb)
+	if err := m.backend.UndoSetupSnap(snapsup.placeInfo(), typ, pb); err != nil {
+		return err
+	}
+
+	st.Lock()
+	defer st.Unlock()
+
+	otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+	if err != nil {
+		return err
+	}
+
+	return m.backend.RemoveSnapDir(snapsup.placeInfo(), otherInstances)
 }
 
 func (m *SnapManager) doUnlinkCurrentSnap(t *state.Task, _ *tomb.Tomb) error {
@@ -445,7 +639,7 @@ func (m *SnapManager) doUnlinkCurrentSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	// Make a copy of configuration of given snap revision
-	if err = config.SaveRevisionConfig(st, snapsup.Name(), snapst.Current); err != nil {
+	if err = config.SaveRevisionConfig(st, snapsup.InstanceName(), snapst.Current); err != nil {
 		return err
 	}
 
@@ -458,7 +652,7 @@ func (m *SnapManager) doUnlinkCurrentSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	// mark as inactive
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 	return nil
 }
 
@@ -477,14 +671,19 @@ func (m *SnapManager) undoUnlinkCurrentSnap(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
+	model, err := Model(st)
+	if err != nil && err != state.ErrNoState {
+		return err
+	}
+
 	snapst.Active = true
-	err = m.backend.LinkSnap(oldInfo)
+	err = m.backend.LinkSnap(oldInfo, model)
 	if err != nil {
 		return err
 	}
 
 	// mark as active again
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 
 	// if we just put back a previous a core snap, request a restart
 	// so that we switch executing its snapd
@@ -494,14 +693,15 @@ func (m *SnapManager) undoUnlinkCurrentSnap(t *state.Task, _ *tomb.Tomb) error {
 }
 
 func (m *SnapManager) doCopySnapData(t *state.Task, _ *tomb.Tomb) error {
-	t.State().Lock()
+	st := t.State()
+	st.Lock()
 	snapsup, snapst, err := snapSetupAndState(t)
-	t.State().Unlock()
+	st.Unlock()
 	if err != nil {
 		return err
 	}
 
-	newInfo, err := readInfo(snapsup.Name(), snapsup.SideInfo)
+	newInfo, err := readInfo(snapsup.InstanceName(), snapsup.SideInfo, 0)
 	if err != nil {
 		return err
 	}
@@ -512,18 +712,42 @@ func (m *SnapManager) doCopySnapData(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	pb := NewTaskProgressAdapterUnlocked(t)
-	return m.backend.CopySnapData(newInfo, oldInfo, pb)
+	if copyDataErr := m.backend.CopySnapData(newInfo, oldInfo, pb); copyDataErr != nil {
+		if oldInfo != nil {
+			// there is another revision of the snap, cannot remove
+			// shared data directory
+			return copyDataErr
+		}
+
+		// cleanup shared snap data directory
+		st.Lock()
+		defer st.Unlock()
+
+		otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+		if err != nil {
+			t.Errorf("cannot undo partial snap %q data copy: %v", snapsup.InstanceName(), err)
+			return copyDataErr
+		}
+		// no other instances of this snap, shared data directory can be
+		// removed now too
+		if err := m.backend.RemoveSnapDataDir(newInfo, otherInstances); err != nil {
+			t.Errorf("cannot undo partial snap %q data copy, failed removing shared directory: %v", snapsup.InstanceName(), err)
+		}
+		return copyDataErr
+	}
+	return nil
 }
 
 func (m *SnapManager) undoCopySnapData(t *state.Task, _ *tomb.Tomb) error {
-	t.State().Lock()
+	st := t.State()
+	st.Lock()
 	snapsup, snapst, err := snapSetupAndState(t)
-	t.State().Unlock()
+	st.Unlock()
 	if err != nil {
 		return err
 	}
 
-	newInfo, err := readInfo(snapsup.Name(), snapsup.SideInfo)
+	newInfo, err := readInfo(snapsup.InstanceName(), snapsup.SideInfo, 0)
 	if err != nil {
 		return err
 	}
@@ -534,7 +758,29 @@ func (m *SnapManager) undoCopySnapData(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	pb := NewTaskProgressAdapterUnlocked(t)
-	return m.backend.UndoCopySnapData(newInfo, oldInfo, pb)
+	if err := m.backend.UndoCopySnapData(newInfo, oldInfo, pb); err != nil {
+		return err
+	}
+
+	if oldInfo != nil {
+		// there is other revision of this snap, cannot remove shared
+		// directory anyway
+		return nil
+	}
+
+	st.Lock()
+	defer st.Unlock()
+
+	otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+	if err != nil {
+		return err
+	}
+	// no other instances of this snap and no other revisions, shared data
+	// directory can be removed
+	if err := m.backend.RemoveSnapDataDir(newInfo, otherInstances); err != nil {
+		return err
+	}
+	return nil
 }
 
 func (m *SnapManager) cleanupCopySnapData(t *state.Task, _ *tomb.Tomb) error {
@@ -562,6 +808,27 @@ func (m *SnapManager) cleanupCopySnapData(t *state.Task, _ *tomb.Tomb) error {
 	return nil
 }
 
+// writeSeqFile writes the sequence file for failover handling
+func writeSeqFile(name string, snapst *SnapState) error {
+	p := filepath.Join(dirs.SnapSeqDir, name+".json")
+	if err := os.MkdirAll(filepath.Dir(p), 0755); err != nil {
+		return err
+	}
+
+	b, err := json.Marshal(&struct {
+		Sequence []*snap.SideInfo `json:"sequence"`
+		Current  string           `json:"current"`
+	}{
+		Sequence: snapst.Sequence,
+		Current:  snapst.Current.String(),
+	})
+	if err != nil {
+		return err
+	}
+
+	return osutil.AtomicWriteFile(p, b, 0644, 0)
+}
+
 func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	st := t.State()
 	st.Lock()
@@ -605,13 +872,27 @@ func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	if snapsup.Required { // set only on install and left alone on refresh
 		snapst.Required = true
 	}
-	// only set userID if unset in snapst and if we actually have
-	// an associated user
-	if snapst.UserID == 0 && snapsup.UserID > 0 {
-		snapst.UserID = snapsup.UserID
+	// only set userID if unset or logged out in snapst and if we
+	// actually have an associated user
+	if snapsup.UserID > 0 {
+		var user *auth.UserState
+		if snapst.UserID != 0 {
+			user, err = auth.User(st, snapst.UserID)
+			if err != nil && err != auth.ErrInvalidUser {
+				return err
+			}
+		}
+		if user == nil {
+			// if the original user installing the snap is
+			// no longer available transfer to user who
+			// triggered this change
+			snapst.UserID = snapsup.UserID
+		}
 	}
+	// keep instance key
+	snapst.InstanceKey = snapsup.InstanceKey
 
-	newInfo, err := readInfo(snapsup.Name(), cand)
+	newInfo, err := readInfo(snapsup.InstanceName(), cand, 0)
 	if err != nil {
 		return err
 	}
@@ -620,12 +901,13 @@ func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	snapst.SetType(newInfo.Type)
 
 	// XXX: this block is slightly ugly, find a pattern when we have more examples
-	err = m.backend.LinkSnap(newInfo)
+	model, _ := Model(st)
+	err = m.backend.LinkSnap(newInfo, model)
 	if err != nil {
 		pb := NewTaskProgressAdapterLocked(t)
 		err := m.backend.UnlinkSnap(newInfo, pb)
 		if err != nil {
-			t.Errorf("cannot cleanup failed attempt at making snap %q available to the system: %v", snapsup.Name(), err)
+			t.Errorf("cannot cleanup failed attempt at making snap %q available to the system: %v", snapsup.InstanceName(), err)
 		}
 	}
 	if err != nil {
@@ -634,17 +916,16 @@ func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 
 	// Restore configuration of the target revision (if available) on revert
 	if snapsup.Revert {
-		if err = config.RestoreRevisionConfig(st, snapsup.Name(), snapsup.Revision()); err != nil {
+		if err = config.RestoreRevisionConfig(st, snapsup.InstanceName(), snapsup.Revision()); err != nil {
 			return err
 		}
 	}
 
 	if len(snapst.Sequence) == 1 {
-		if err := m.createSnapCookie(st, snapsup.Name()); err != nil {
+		if err := m.createSnapCookie(st, snapsup.InstanceName()); err != nil {
 			return fmt.Errorf("cannot create snap cookie: %v", err)
 		}
 	}
-
 	// save for undoLinkSnap
 	t.Set("old-trymode", oldTryMode)
 	t.Set("old-devmode", oldDevMode)
@@ -655,7 +936,37 @@ func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	t.Set("old-current", oldCurrent)
 	t.Set("old-candidate-index", oldCandidateIndex)
 	// Do at the end so we only preserve the new state if it worked.
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
+	// write sequence file for failover helpers
+	if err := writeSeqFile(snapsup.InstanceName(), snapst); err != nil {
+		return err
+	}
+
+	// Compatibility with old snapd: check if we have auto-connect task and
+	// if not, inject it after self (link-snap) for snaps that are not core
+	if newInfo.Type != snap.TypeOS {
+		var hasAutoConnect, hasSetupProfiles bool
+		for _, other := range t.Change().Tasks() {
+			// Check if this is auto-connect task for same snap and we it's part of the change with setup-profiles task
+			if other.Kind() == "auto-connect" || other.Kind() == "setup-profiles" {
+				otherSnapsup, err := TaskSnapSetup(other)
+				if err != nil {
+					return err
+				}
+				if snapsup.InstanceName() == otherSnapsup.InstanceName() {
+					if other.Kind() == "auto-connect" {
+						hasAutoConnect = true
+					} else {
+						hasSetupProfiles = true
+					}
+				}
+			}
+		}
+		if !hasAutoConnect && hasSetupProfiles {
+			InjectAutoConnect(t, snapsup)
+		}
+	}
+
 	// Make sure if state commits and snapst is mutated we won't be rerun
 	t.SetStatus(state.DoneStatus)
 
@@ -666,17 +977,39 @@ func (m *SnapManager) doLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	return nil
 }
 
-// maybeRestart will schedule a reboot or restart as needed for the just linked
-// snap with info if it's a core or kernel snap.
+// maybeRestart will schedule a reboot or restart as needed for the
+// just linked snap with info if it's a core or snapd or kernel snap.
 func maybeRestart(t *state.Task, info *snap.Info) {
 	st := t.State()
-	if release.OnClassic && info.Type == snap.TypeOS {
-		t.Logf("Requested daemon restart.")
-		st.RequestRestart(state.RestartDaemon)
+
+	// TODO: once classic uses the snapd snap we need to restart
+	//       here too
+	if release.OnClassic {
+		if info.Type == snap.TypeOS {
+			t.Logf("Requested daemon restart.")
+			st.RequestRestart(state.RestartDaemon)
+		}
+		return
 	}
-	if !release.OnClassic && boot.KernelOrOsRebootRequired(info) {
+
+	// On a core system we may need a full reboot if
+	// core/base or the kernel changes.
+	if boot.ChangeRequiresReboot(info) {
 		t.Logf("Requested system restart.")
 		st.RequestRestart(state.RestartSystem)
+		return
+	}
+
+	// On core systems that use a base snap we need to restart
+	// snapd when the snapd snap changes.
+	model, err := Model(st)
+	if err != nil {
+		logger.Noticef("cannot get model assertion: %v", model)
+		return
+	}
+	if model.Base() != "" && info.InstanceName() == "snapd" {
+		t.Logf("Requested daemon restart (snapd snap).")
+		st.RequestRestart(state.RestartDaemon)
 	}
 }
 
@@ -731,7 +1064,12 @@ func (m *SnapManager) undoLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	if len(snapst.Sequence) == 1 {
-		if err := m.removeSnapCookie(st, snapsup.Name()); err != nil {
+		err = m.backend.DiscardSnapNamespace(snapsup.InstanceName())
+		if err != nil {
+			t.Errorf("cannot discard snap namespace %q, will retry in 3 mins: %s", snapsup.InstanceName(), err)
+			return &state.Retry{After: 3 * time.Minute}
+		}
+		if err := m.removeSnapCookie(st, snapsup.InstanceName()); err != nil {
 			return fmt.Errorf("cannot remove snap cookie: %v", err)
 		}
 	}
@@ -760,13 +1098,13 @@ func (m *SnapManager) undoLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	snapst.JailMode = oldJailMode
 	snapst.Classic = oldClassic
 
-	newInfo, err := readInfo(snapsup.Name(), snapsup.SideInfo)
+	newInfo, err := readInfo(snapsup.InstanceName(), snapsup.SideInfo, 0)
 	if err != nil {
 		return err
 	}
 
 	if len(snapst.Sequence) == 1 {
-		if err = config.RestoreRevisionConfig(st, snapsup.Name(), oldCurrent); err != nil {
+		if err = config.RestoreRevisionConfig(st, snapsup.InstanceName(), oldCurrent); err != nil {
 			return err
 		}
 	}
@@ -777,7 +1115,11 @@ func (m *SnapManager) undoLinkSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	// mark as inactive
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
+	// write sequence file for failover helpers
+	if err := writeSeqFile(snapsup.InstanceName(), snapst); err != nil {
+		return err
+	}
 	// Make sure if state commits and snapst is mutated we won't be rerun
 	t.SetStatus(state.UndoneStatus)
 
@@ -812,7 +1154,7 @@ func (m *SnapManager) doSwitchSnapChannel(t *state.Task, _ *tomb.Tomb) error {
 		snapst.CurrentSideInfo().Channel = snapsup.Channel
 	}
 
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 	return nil
 }
 
@@ -827,7 +1169,7 @@ func (m *SnapManager) doSwitchSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 	snapst.Channel = snapsup.Channel
 
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 	return nil
 }
 
@@ -844,7 +1186,7 @@ func (m *SnapManager) doToggleSnapFlags(t *state.Task, _ *tomb.Tomb) error {
 	// for now we support toggling only ignore-validation
 	snapst.IgnoreValidation = snapsup.IgnoreValidation
 
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 	return nil
 }
 
@@ -867,9 +1209,14 @@ func (m *SnapManager) startSnapServices(t *state.Task, _ *tomb.Tomb) error {
 		return nil
 	}
 
+	startupOrdered, err := snap.SortServices(svcs)
+	if err != nil {
+		return err
+	}
+
 	pb := NewTaskProgressAdapterUnlocked(t)
 	st.Unlock()
-	err = m.backend.StartServices(svcs, pb)
+	err = m.backend.StartServices(startupOrdered, pb)
 	st.Lock()
 	return err
 }
@@ -893,9 +1240,14 @@ func (m *SnapManager) stopSnapServices(t *state.Task, _ *tomb.Tomb) error {
 		return nil
 	}
 
+	var stopReason snap.ServiceStopReason
+	if err := t.Get("stop-reason", &stopReason); err != nil && err != state.ErrNoState {
+		return err
+	}
+
 	pb := NewTaskProgressAdapterUnlocked(t)
 	st.Unlock()
-	err = m.backend.StopServices(svcs, pb)
+	err = m.backend.StopServices(svcs, stopReason, pb)
 	st.Lock()
 	return err
 }
@@ -911,7 +1263,7 @@ func (m *SnapManager) doUnlinkSnap(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	info, err := Info(t.State(), snapsup.Name(), snapsup.Revision())
+	info, err := Info(t.State(), snapsup.InstanceName(), snapsup.Revision())
 	if err != nil {
 		return err
 	}
@@ -924,22 +1276,23 @@ func (m *SnapManager) doUnlinkSnap(t *state.Task, _ *tomb.Tomb) error {
 
 	// mark as inactive
 	snapst.Active = false
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 
 	return err
 }
 
 func (m *SnapManager) doClearSnapData(t *state.Task, _ *tomb.Tomb) error {
-	t.State().Lock()
+	st := t.State()
+	st.Lock()
 	snapsup, snapst, err := snapSetupAndState(t)
-	t.State().Unlock()
+	st.Unlock()
 	if err != nil {
 		return err
 	}
 
-	t.State().Lock()
-	info, err := Info(t.State(), snapsup.Name(), snapsup.Revision())
-	t.State().Unlock()
+	st.Lock()
+	info, err := Info(t.State(), snapsup.InstanceName(), snapsup.Revision())
+	st.Unlock()
 	if err != nil {
 		return err
 	}
@@ -948,11 +1301,23 @@ func (m *SnapManager) doClearSnapData(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	// Only remove data common between versions if this is the last version
 	if len(snapst.Sequence) == 1 {
+		// Only remove data common between versions if this is the last version
 		if err = m.backend.RemoveSnapCommonData(info); err != nil {
 			return err
 		}
+
+		st.Lock()
+		defer st.Unlock()
+
+		otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+		if err != nil {
+			return err
+		}
+		// Snap data directory can be removed now too
+		if err := m.backend.RemoveSnapDataDir(info, otherInstances); err != nil {
+			return err
+		}
 	}
 
 	return nil
@@ -969,7 +1334,7 @@ func (m *SnapManager) doDiscardSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	if snapst.Current == snapsup.Revision() && snapst.Active {
-		return fmt.Errorf("internal error: cannot discard snap %q: still active", snapsup.Name())
+		return fmt.Errorf("internal error: cannot discard snap %q: still active", snapsup.InstanceName())
 	}
 
 	if len(snapst.Sequence) == 1 {
@@ -997,28 +1362,37 @@ func (m *SnapManager) doDiscardSnap(t *state.Task, _ *tomb.Tomb) error {
 	}
 	err = m.backend.RemoveSnapFiles(snapsup.placeInfo(), typ, pb)
 	if err != nil {
-		t.Errorf("cannot remove snap file %q, will retry in 3 mins: %s", snapsup.Name(), err)
+		t.Errorf("cannot remove snap file %q, will retry in 3 mins: %s", snapsup.InstanceName(), err)
 		return &state.Retry{After: 3 * time.Minute}
 	}
 	if len(snapst.Sequence) == 0 {
 		// Remove configuration associated with this snap.
-		err = config.DeleteSnapConfig(st, snapsup.Name())
+		err = config.DeleteSnapConfig(st, snapsup.InstanceName())
 		if err != nil {
 			return err
 		}
-		err = m.backend.DiscardSnapNamespace(snapsup.Name())
+		err = m.backend.DiscardSnapNamespace(snapsup.InstanceName())
 		if err != nil {
-			t.Errorf("cannot discard snap namespace %q, will retry in 3 mins: %s", snapsup.Name(), err)
+			t.Errorf("cannot discard snap namespace %q, will retry in 3 mins: %s", snapsup.InstanceName(), err)
 			return &state.Retry{After: 3 * time.Minute}
 		}
-		if err := m.removeSnapCookie(st, snapsup.Name()); err != nil {
+		if err := m.removeSnapCookie(st, snapsup.InstanceName()); err != nil {
 			return fmt.Errorf("cannot remove snap cookie: %v", err)
 		}
+
+		otherInstances, err := hasOtherInstances(st, snapsup.InstanceName())
+		if err != nil {
+			return err
+		}
+
+		if err := m.backend.RemoveSnapDir(snapsup.placeInfo(), otherInstances); err != nil {
+			return fmt.Errorf("cannot remove snap directory: %v", err)
+		}
 	}
-	if err = config.DiscardRevisionConfig(st, snapsup.Name(), snapsup.Revision()); err != nil {
+	if err = config.DiscardRevisionConfig(st, snapsup.InstanceName(), snapsup.Revision()); err != nil {
 		return err
 	}
-	Set(st, snapsup.Name(), snapst)
+	Set(st, snapsup.InstanceName(), snapst)
 	return nil
 }
 
@@ -1068,7 +1442,7 @@ func (m *SnapManager) doSetAutoAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	curInfo, err := snapst.CurrentInfo()
 	if err != nil {
 		return err
@@ -1107,7 +1481,7 @@ func (m *SnapManager) doRemoveAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 
 	err = m.backend.RemoveSnapAliases(snapName)
 	if err != nil {
@@ -1127,7 +1501,7 @@ func (m *SnapManager) doSetupAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	curAliases := snapst.Aliases
 
 	_, _, err = applyAliasesChange(snapName, autoDis, nil, snapst.AutoAliasesDisabled, curAliases, m.backend, doApply)
@@ -1148,7 +1522,7 @@ func (m *SnapManager) doRefreshAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	curInfo, err := snapst.CurrentInfo()
 	if err != nil {
 		return err
@@ -1194,7 +1568,7 @@ func (m *SnapManager) undoRefreshAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	curAutoDisabled := snapst.AutoAliasesDisabled
 	autoDisabled := curAutoDisabled
 	if err = t.Get("old-auto-aliases-disabled", &autoDisabled); err != nil && err != state.ErrNoState {
@@ -1283,12 +1657,12 @@ func (m *SnapManager) undoRefreshAliases(t *state.Task, _ *tomb.Tomb) error {
 		}
 	}
 
-	for snapName, snapst := range newSnapStates {
-		if conflicting[snapName] {
+	for instanceName, snapst := range newSnapStates {
+		if conflicting[instanceName] {
 			// keep as it was
 			continue
 		}
-		Set(st, snapName, snapst)
+		Set(st, instanceName, snapst)
 	}
 	return nil
 }
@@ -1306,7 +1680,7 @@ func (m *SnapManager) doPruneAutoAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	autoDisabled := snapst.AutoAliasesDisabled
 	curAliases := snapst.Aliases
 
@@ -1385,7 +1759,7 @@ func (m *SnapManager) doAlias(t *state.Task, _ *tomb.Tomb) error {
 		return err
 	}
 
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 	curInfo, err := snapst.CurrentInfo()
 	if err != nil {
 		return err
@@ -1424,7 +1798,7 @@ func (m *SnapManager) doDisableAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 
 	oldAutoDisabled := snapst.AutoAliasesDisabled
 	oldAliases := snapst.Aliases
@@ -1459,7 +1833,7 @@ func (m *SnapManager) doUnalias(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	snapName := snapsup.InstanceName()
 
 	autoDisabled := snapst.AutoAliasesDisabled
 	oldAliases := snapst.Aliases
@@ -1500,7 +1874,7 @@ func (m *SnapManager) doPreferAliases(t *state.Task, _ *tomb.Tomb) error {
 	if err != nil {
 		return err
 	}
-	snapName := snapsup.Name()
+	instanceName := snapsup.InstanceName()
 
 	if !snapst.AutoAliasesDisabled {
 		// already enabled, nothing to do
@@ -1508,7 +1882,7 @@ func (m *SnapManager) doPreferAliases(t *state.Task, _ *tomb.Tomb) error {
 	}
 
 	curAliases := snapst.Aliases
-	aliasConflicts, err := checkAliasesConflicts(st, snapName, autoEn, curAliases, nil)
+	aliasConflicts, err := checkAliasesConflicts(st, instanceName, autoEn, curAliases, nil)
 	conflErr, isConflErr := err.(*AliasConflictError)
 	if err != nil && !isConflErr {
 		return err
@@ -1518,7 +1892,7 @@ func (m *SnapManager) doPreferAliases(t *state.Task, _ *tomb.Tomb) error {
 		return conflErr
 	}
 	// proceed to disable conflicting aliases as needed
-	// before re-enabling snapName aliases
+	// before re-enabling instanceName aliases
 
 	otherSnapStates := make(map[string]*SnapState, len(aliasConflicts))
 	otherSnapDisabled := make(map[string]*otherDisabledAliases, len(aliasConflicts))
@@ -1552,7 +1926,7 @@ func (m *SnapManager) doPreferAliases(t *state.Task, _ *tomb.Tomb) error {
 		otherSnapStates[otherSnap] = &otherSnapState
 	}
 
-	added, removed, err := applyAliasesChange(snapName, autoDis, curAliases, autoEn, curAliases, m.backend, snapst.AliasesPending)
+	added, removed, err := applyAliasesChange(instanceName, autoDis, curAliases, autoEn, curAliases, m.backend, snapst.AliasesPending)
 	if err != nil {
 		return err
 	}
@@ -1569,6 +1943,42 @@ func (m *SnapManager) doPreferAliases(t *state.Task, _ *tomb.Tomb) error {
 	t.Set("old-auto-aliases-disabled", true)
 	t.Set("old-aliases-v2", curAliases)
 	snapst.AutoAliasesDisabled = false
-	Set(st, snapName, snapst)
+	Set(st, instanceName, snapst)
 	return nil
 }
+
+// InjectTasks makes all the halt tasks of the mainTask wait for extraTasks;
+// extraTasks join the same lane and change as the mainTask.
+func InjectTasks(mainTask *state.Task, extraTasks *state.TaskSet) {
+	lanes := mainTask.Lanes()
+	if len(lanes) == 1 && lanes[0] == 0 {
+		lanes = nil
+	}
+	for _, l := range lanes {
+		extraTasks.JoinLane(l)
+	}
+
+	chg := mainTask.Change()
+	// Change shouldn't normally be nil, except for cases where
+	// this helper is used before tasks are added to a change.
+	if chg != nil {
+		chg.AddAll(extraTasks)
+	}
+
+	// make all halt tasks of the mainTask wait on extraTasks
+	ht := mainTask.HaltTasks()
+	for _, t := range ht {
+		t.WaitAll(extraTasks)
+	}
+
+	// make the extra tasks wait for main task
+	extraTasks.WaitFor(mainTask)
+}
+
+func InjectAutoConnect(mainTask *state.Task, snapsup *SnapSetup) {
+	st := mainTask.State()
+	autoConnect := st.NewTask("auto-connect", fmt.Sprintf(i18n.G("Automatically connect eligible plugs and slots of snap %q"), snapsup.InstanceName()))
+	autoConnect.Set("snap-setup", snapsup)
+	InjectTasks(mainTask, state.NewTaskSet(autoConnect))
+	mainTask.Logf("added auto-connect task")
+}
