@@ -44,3 +44,24 @@ func MockTaskTimes(t *Task, spawnTime, readyTime time.Time) {
 	t.spawnTime = spawnTime
 	t.readyTime = readyTime
 }
+
+func (s *State) AddWarning(message string, lastAdded, lastShown time.Time, expireAfter, repeatAfter time.Duration) {
+	s.addWarning(Warning{
+		message:     message,
+		lastShown:   lastShown,
+		expireAfter: expireAfter,
+		repeatAfter: repeatAfter,
+	}, lastAdded)
+}
+
+func (w Warning) LastAdded() time.Time {
+	return w.lastAdded
+}
+
+var (
+	ErrNoWarningMessage     = errNoWarningMessage
+	ErrBadWarningMessage    = errBadWarningMessage
+	ErrNoWarningFirstAdded  = errNoWarningFirstAdded
+	ErrNoWarningExpireAfter = errNoWarningExpireAfter
+	ErrNoWarningRepeatAfter = errNoWarningRepeatAfter
+)
