@@ -4,57 +4,112 @@
 . "$TESTSLIB/quiet.sh"
 
 debian_name_package() {
-    case "$1" in
-        xdelta3|curl|python3-yaml|kpartx|busybox-static|nfs-kernel-server)
-            echo "$1"
-            ;;
-        man)
-            echo "man-db"
-            ;;
-        *)
-            echo "$1"
-            ;;
-    esac
+    for i in "$@"; do
+        case "$i" in
+            man)
+                echo "man-db"
+                ;;
+            *)
+                echo "$i"
+                ;;
+        esac
+    done
 }
 
 ubuntu_14_04_name_package() {
-    case "$1" in
-        printer-driver-cups-pdf)
-            echo "cups-pdf"
-            ;;
-        *)
-            debian_name_package "$1"
-            ;;
-    esac
+    for i in "$@"; do
+        case "$i" in
+            printer-driver-cups-pdf)
+                echo "cups-pdf"
+                ;;
+            *)
+                debian_name_package "$i"
+                ;;
+        esac
+    done
 }
 
 fedora_name_package() {
+    for i in "$@"; do
+        case "$i" in
+            openvswitch-switch)
+                echo "openvswitch"
+                ;;
+            printer-driver-cups-pdf)
+                echo "cups-pdf"
+                ;;
+            python3-gi)
+                echo "python3-gobject"
+                ;;
+            *)
+                echo "$i"
+                ;;
+        esac
+    done
+}
+
+amazon_name_package() {
+    for i in "$@"; do
+        case "$i" in
+            xdelta3)
+                echo "xdelta"
+                ;;
+            *)
+                echo "$i"
+                ;;
+        esac
+    done
+}
+
+opensuse_name_package() {
+    for i in "$@"; do
+        case "$i" in
+            python3-yaml)
+                echo "python3-PyYAML"
+                ;;
+            dbus-x11)
+                echo "dbus-1-x11"
+                ;;
+            printer-driver-cups-pdf)
+                echo "cups-pdf"
+                ;;
+            python3-dbus)
+                # In OpenSUSE Leap 15, this is renamed to python3-dbus-python
+                echo "dbus-1-python3"
+                ;;
+            python3-gi)
+                echo "python3-gobject"
+                ;;
+            *)
+                echo "$i"
+                ;;
+        esac
+    done
+}
+
+arch_name_package() {
     case "$1" in
-        xdelta3|jq|curl|python3-yaml)
-            echo "$1"
+        python3-yaml)
+            echo "python-yaml"
+            ;;
+        dbus-x11)
+            # no separate dbus-x11 package in arch
+            echo "dbus"
+            ;;
+        printer-driver-cups-pdf)
+            echo "cups-pdf"
             ;;
         openvswitch-switch)
             echo "openvswitch"
             ;;
-        printer-driver-cups-pdf)
-            echo "cups-pdf"
+        man)
+            echo "man-db"
             ;;
-        *)
-            echo "$1"
+        python3-dbus)
+            echo "python-dbus"
             ;;
-    esac
-}
-
-opensuse_name_package() {
-    case "$1" in
-        python3-yaml)
-            echo "python3-PyYAML"
-            ;;
-        dbus-x11)
-            echo "dbus-1-x11"
-            ;;
-        printer-driver-cups-pdf)
-            echo "cups-pdf"
+        python3-gi)
+            echo "python-gobject"
             ;;
         *)
             echo "$1"
@@ -65,16 +120,22 @@ opensuse_name_package() {
 distro_name_package() {
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*)
-            ubuntu_14_04_name_package "$1"
+            ubuntu_14_04_name_package "$@"
             ;;
         ubuntu-*|debian-*)
-            debian_name_package "$1"
+            debian_name_package "$@"
             ;;
         fedora-*)
-            fedora_name_package "$1"
+            fedora_name_package "$@"
+            ;;
+        amazon-*|centos-*)
+            amazon_name_package "$@"
             ;;
         opensuse-*)
-            opensuse_name_package "$1"
+            opensuse_name_package "$@"
+            ;;
+        arch-*)
+            arch_name_package "$1"
             ;;
         *)
             echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
@@ -111,10 +172,16 @@ distro_install_local_package() {
             apt install $flags "$@"
             ;;
         fedora-*)
-            dnf -q -y install "$@"
+            quiet dnf -y install --setopt=install_weak_deps=False "$@"
+            ;;
+        amazon-*|centos-*)
+            quiet yum -y localinstall "$@"
             ;;
         opensuse-*)
-            zypper -q install -y "$@"
+            quiet rpm -i "$@"
+            ;;
+        arch-*)
+            pacman -U --noconfirm "$@"
             ;;
         *)
             echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
@@ -124,11 +191,27 @@ distro_install_local_package() {
 }
 
 distro_install_package() {
+    orig_xtrace=$(set -o | awk '/xtrace / { print $2 }')
+    set +x
+    echo "distro_install_package $*"
     # Parse additional arguments; once we find the first unknown
     # part we break argument parsing and process all further
     # arguments as package names.
     APT_FLAGS=
     DNF_FLAGS=
+    if [[ "$SPREAD_SYSTEM" == fedora-* ]]; then
+        # Fedora images we use come with a number of preinstalled package, among
+        # them gtk3. Those packages are needed to run the tests. The
+        # xdg-desktop-portal-gtk package uses this in the spec:
+        #
+        #   Supplements:    (gtk3 and (flatpak or snapd))
+        #
+        # As a result, when snapd is installed, we will unintentionally pull in
+        # xdg-desktop-portal-gtk and its dependencies breaking tests. For this
+        # reason, disable weak deps altogether.
+        DNF_FLAGS="--setopt=install_weak_deps=False"
+    fi
+    YUM_FLAGS=
     ZYPPER_FLAGS=
     while [ -n "$1" ]; do
         case "$1" in
@@ -136,6 +219,7 @@ distro_install_package() {
                 APT_FLAGS="$APT_FLAGS --no-install-recommends"
                 DNF_FLAGS="$DNF_FLAGS --setopt=install_weak_deps=False"
                 ZYPPER_FLAGS="$ZYPPER_FLAGS --no-recommends"
+                # TODO no way to set this for yum?
                 shift
                 ;;
             *)
@@ -164,61 +248,84 @@ distro_install_package() {
         ;;
     esac
 
-    for pkg in "$@" ; do
-        package_name=$(distro_name_package "$pkg")
-        # When we could not find a different package name for the distribution
-        # we're running on we try the package name given as last attempt
-        if [ -z "$package_name" ]; then
-            package_name="$pkg"
-        fi
+    # shellcheck disable=SC2207
+    pkg_names=($(
+        for pkg in "$@" ; do
+            package_name=$(distro_name_package "$pkg")
+            # When we could not find a different package name for the distribution
+            # we're running on we try the package name given as last attempt
+            if [ -z "$package_name" ]; then
+                package_name="$pkg"
+            fi
+            echo "$package_name"
+        done
+    ))
 
-        case "$SPREAD_SYSTEM" in
-            ubuntu-*|debian-*)
-                # shellcheck disable=SC2086
-                quiet apt-get install $APT_FLAGS -y "$package_name"
-                ;;
-            fedora-*)
-                # shellcheck disable=SC2086
-                dnf -q -y --refresh install $DNF_FLAGS "$package_name"
-                ;;
-            opensuse-*)
-                # shellcheck disable=SC2086
-                zypper -q install -y $ZYPPER_FLAGS "$package_name"
-                ;;
-            *)
-                echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
-                exit 1
-                ;;
-        esac
-    done
+    case "$SPREAD_SYSTEM" in
+        ubuntu-*|debian-*)
+            # shellcheck disable=SC2086
+            quiet apt-get install $APT_FLAGS -y "${pkg_names[@]}"
+            ;;
+        fedora-*)
+            # shellcheck disable=SC2086
+            quiet dnf -y --refresh install $DNF_FLAGS "${pkg_names[@]}"
+            ;;
+        amazon-*|centos-*)
+            # shellcheck disable=SC2086
+            quiet yum -y install $YUM_FLAGS "${pkg_names[@]}"
+            ;;
+        opensuse-*)
+            # shellcheck disable=SC2086
+            quiet zypper install -y $ZYPPER_FLAGS "${pkg_names[@]}"
+            ;;
+        arch-*)
+            # shellcheck disable=SC2086
+            pacman -Suq --needed --noconfirm "${pkg_names[@]}"
+            ;;
+        *)
+            echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
+            exit 1
+            ;;
+    esac
+    test "$orig_xtrace" = on && set -x
 }
 
 distro_purge_package() {
-    for pkg in "$@" ; do
-        package_name=$(distro_name_package "$pkg")
-        # When we could not find a different package name for the distribution
-        # we're running on we try the package name given as last attempt
-        if [ -z "$package_name" ]; then
-            package_name="$pkg"
-        fi
+    # shellcheck disable=SC2046
+    set -- $(
+        for pkg in "$@" ; do
+            package_name=$(distro_name_package "$pkg")
+            # When we could not find a different package name for the distribution
+            # we're running on we try the package name given as last attempt
+            if [ -z "$package_name" ]; then
+                package_name="$pkg"
+            fi
+            echo "$package_name"
+        done
+        )
 
-        case "$SPREAD_SYSTEM" in
-            ubuntu-*|debian-*)
-                quiet apt-get remove -y --purge -y "$package_name"
-                ;;
-            fedora-*)
-                dnf -y -q remove "$package_name"
-                dnf -q clean all
-                ;;
-            opensuse-*)
-                zypper -q remove -y "$package_name"
-                ;;
-            *)
-                echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
-                exit 1
-                ;;
-        esac
-    done
+    case "$SPREAD_SYSTEM" in
+        ubuntu-*|debian-*)
+            quiet apt-get remove -y --purge -y "$@"
+            ;;
+        fedora-*)
+            quiet dnf -y remove "$@"
+            quiet dnf clean all
+            ;;
+        amazon-*|centos-*)
+            quiet yum -y remove "$@"
+            ;;
+        opensuse-*)
+            quiet zypper remove -y "$@"
+            ;;
+        arch-*)
+            pacman -Rnsc --noconfirm "$@"
+            ;;
+        *)
+            echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
+            exit 1
+            ;;
+    esac
 }
 
 distro_update_package_db() {
@@ -227,10 +334,18 @@ distro_update_package_db() {
             quiet apt-get update
             ;;
         fedora-*)
-            dnf -q makecache
+            quiet dnf clean all
+            quiet dnf makecache
+            ;;
+        amazon-*|centos-*)
+            quiet yum clean all
+            quiet yum makecache
             ;;
         opensuse-*)
-            zypper -q refresh
+            quiet zypper --gpg-auto-import-keys refresh
+            ;;
+        arch-*)
+            pacman -Syq
             ;;
         *)
             echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
@@ -244,9 +359,18 @@ distro_clean_package_cache() {
         ubuntu-*|debian-*)
             quiet apt-get clean
             ;;
+        fedora-*)
+            dnf clean all
+            ;;
+        amazon-*|centos-*)
+            yum clean all
+            ;;
         opensuse-*)
             zypper -q clean --all
             ;;
+        arch-*)
+            pacman -Sccq --noconfirm
+            ;;
         *)
             echo "ERROR: Unsupported distribution $SPREAD_SYSTEM"
             exit 1
@@ -260,10 +384,15 @@ distro_auto_remove_packages() {
             quiet apt-get -y autoremove
             ;;
         fedora-*)
-            dnf -q -y autoremove
+            quiet dnf -y autoremove
+            ;;
+        amazon-*|centos-*)
+            quiet yum -y autoremove
             ;;
         opensuse-*)
             ;;
+        arch-*)
+            ;;
         *)
             echo "ERROR: Unsupported distribution '$SPREAD_SYSTEM'"
             exit 1
@@ -279,9 +408,19 @@ distro_query_package_info() {
         fedora-*)
             dnf info "$1"
             ;;
+        amazon-*|centos-*)
+            yum info "$1"
+            ;;
         opensuse-*)
             zypper info "$1"
             ;;
+        arch-*)
+            pacman -Si "$1"
+            ;;
+        *)
+            echo "ERROR: Unsupported distribution '$SPREAD_SYSTEM'"
+            exit 1
+            ;;
     esac
 }
 
@@ -306,14 +445,18 @@ distro_install_build_snapd(){
                 # shellcheck disable=SC2125
                 packages="${GOHOME}"/snapd_*.deb
                 ;;
-            fedora-*)
+            fedora-*|amazon-*|centos-*)
                 # shellcheck disable=SC2125
-                packages="${GOHOME}"/snap-confine*.rpm\ "${GOPATH}"/snapd*.rpm
+                packages="${GOHOME}"/snap-confine*.rpm\ "${GOPATH%%:*}"/snapd*.rpm
                 ;;
             opensuse-*)
                 # shellcheck disable=SC2125
                 packages="${GOHOME}"/snapd*.rpm
                 ;;
+            arch-*)
+                # shellcheck disable=SC2125
+                packages="${GOHOME}"/snapd*.pkg.tar.xz
+                ;;
             *)
                 exit 1
                 ;;
@@ -322,6 +465,17 @@ distro_install_build_snapd(){
         # shellcheck disable=SC2086
         distro_install_local_package $packages
 
+        if [[ "$SPREAD_SYSTEM" == arch-* ]]; then
+            # Arch policy does not allow calling daemon-reloads in package
+            # install scripts
+            systemctl daemon-reload
+
+            # AppArmor policy needs to be reloaded
+            if systemctl show -p ActiveState apparmor.service | MATCH 'ActiveState=active'; then
+                systemctl restart apparmor.service
+            fi
+        fi
+
         # On some distributions the snapd.socket is not yet automatically
         # enabled as we don't have a systemd present configuration approved
         # by the distribution for it in place yet.
@@ -338,9 +492,13 @@ distro_get_package_extension() {
         ubuntu-*|debian-*)
             echo "deb"
             ;;
-        fedora-*|opensuse-*)
+        fedora-*|opensuse-*|amazon-*|centos-*)
             echo "rpm"
             ;;
+        arch-*)
+            # default /etc/makepkg.conf setting
+            echo "pkg.tar.xz"
+            ;;
     esac
 }
 
@@ -350,32 +508,40 @@ pkg_dependencies_ubuntu_generic(){
         automake
         autotools-dev
         build-essential
+        clang
         curl
         devscripts
         expect
+        gdb
         gdebi-core
         git
         indent
         jq
+        apparmor-utils
         libapparmor-dev
         libglib2.0-dev
         libseccomp-dev
         libudev-dev
+        man
         netcat-openbsd
         pkg-config
         python3-docutils
-        rng-tools
         udev
+        udisks2
+        upower
+        uuid-runtime
         "
 }
 
 pkg_dependencies_ubuntu_classic(){
     echo "
+        avahi-daemon
         cups
         dbus-x11
         gnome-keyring
         jq
         man
+        nfs-kernel-server
         printer-driver-cups-pdf
         python3-yaml
         upower
@@ -386,81 +552,180 @@ pkg_dependencies_ubuntu_classic(){
 
     case "$SPREAD_SYSTEM" in
         ubuntu-14.04-*)
-            echo "
-                linux-image-extra-$(uname -r)
-                pollinate
-                "
+                pkg_linux_image_extra
             ;;
         ubuntu-16.04-32)
             echo "
-                linux-image-extra-$(uname -r)
-                pollinate
+                gccgo-6
+                evolution-data-server
+                gnome-online-accounts
                 "
+                pkg_linux_image_extra
             ;;
         ubuntu-16.04-64)
             echo "
+                evolution-data-server
                 gccgo-6
+                gnome-online-accounts
                 kpartx
                 libvirt-bin
-                linux-image-extra-$(uname -r)
-                pollinate
                 qemu
                 x11-utils
                 xvfb
                 "
+                pkg_linux_image_extra
+            ;;
+        ubuntu-17.10-64)
+                pkg_linux_image_extra
+            ;;
+        ubuntu-18.04-64)
+            echo "
+                gccgo-8
+                evolution-data-server
+                "
+            ;;
+        ubuntu-18.10-64)
+            echo "
+                evolution-data-server
+                "
             ;;
         ubuntu-*)
             echo "
-                linux-image-extra-$(uname -r)
-                pollinate
+                squashfs-tools
                 "
             ;;
         debian-*)
             echo "
+                evolution-data-server
                 net-tools
                 "
             ;;
     esac
 }
 
+pkg_linux_image_extra (){
+    if apt-cache show "linux-image-extra-$(uname -r)" > /dev/null 2>&1; then
+        echo "linux-image-extra-$(uname -r)";
+    else
+        if apt-cache show "linux-modules-extra-$(uname -r)" > /dev/null 2>&1; then
+            echo "linux-modules-extra-$(uname -r)";
+        else
+            echo "cannot find a matching kernel modules package";
+            exit 1;
+        fi;
+    fi
+}
+
 pkg_dependencies_ubuntu_core(){
     echo "
-        linux-image-extra-$(uname -r)
         pollinate
         "
+        pkg_linux_image_extra
 }
 
 pkg_dependencies_fedora(){
+    echo "
+        clang
+        curl
+        dbus-x11
+        evolution-data-server
+        expect
+        git
+        golang
+        jq
+        iptables-services
+        man
+        mock
+        net-tools
+        nfs-utils
+        python3-yaml
+        redhat-lsb-core
+        rpm-build
+        udisks2
+        xdg-user-dirs
+        xdg-utils
+        strace
+        "
+}
+
+pkg_dependencies_amazon(){
     echo "
         curl
         dbus-x11
         expect
         git
         golang
+        grub2-tools
         jq
+        iptables-services
+        man
         mock
-        redhat-lsb-core
+        nc
+        net-tools
+        nfs-utils
+        system-lsb-core
         rpm-build
         xdg-user-dirs
+        xdg-utils
+        udisks2
         "
 }
 
 pkg_dependencies_opensuse(){
     echo "
+        apparmor-profiles
+        clang
         curl
+        evolution-data-server
         expect
         git
         golang-packaging
         jq
         lsb-release
+        man
+        nfs-kernel-server
+        python3-yaml
         netcat-openbsd
         osc
-        rng-tools
-        xdg-utils
+        udisks2
+        uuidd
         xdg-user-dirs
+        xdg-utils
         "
 }
 
+pkg_dependencies_arch(){
+    echo "
+    base-devel
+    bash-completion
+    clang
+    curl
+    evolution-data-server
+    expect
+    git
+    go
+    go-tools
+    jq
+    libseccomp
+    libcap
+    libx11
+    net-tools
+    nfs-utils
+    openbsd-netcat
+    python
+    python-docutils
+    python3-yaml
+    squashfs-tools
+    shellcheck
+    strace
+    udisks2
+    xdg-user-dirs
+    xdg-utils
+    xfsprogs
+    apparmor
+    "
+}
+
 pkg_dependencies(){
     case "$SPREAD_SYSTEM" in
         ubuntu-core-16-*)
@@ -474,9 +739,15 @@ pkg_dependencies(){
         fedora-*)
             pkg_dependencies_fedora
             ;;
+        amazon-*|centos-*)
+            pkg_dependencies_amazon
+            ;;
         opensuse-*)
             pkg_dependencies_opensuse
             ;;
+        arch-*)
+            pkg_dependencies_arch
+            ;;
         *)
             ;;
     esac
@@ -487,3 +758,34 @@ install_pkg_dependencies(){
     # shellcheck disable=SC2086
     distro_install_package $pkgs
 }
+
+# upgrade distribution and indicate if reboot is needed by outputting 'reboot'
+# to stdout
+distro_upgrade() {
+    case "$SPREAD_SYSTEM" in
+        arch-*)
+            # Arch does not support partial upgrades. On top of this, the image
+            # we are running in may have been built some time ago and we need to
+            # upgrade so that the tests are run with the same package versions
+            # as the users will have. We basically need to run pacman -Syu.
+            # Since there is no way to tell if we can continue after upgrading
+            # (eg. the kernel package or systemd got updated ) issue a reboot
+            # instead.
+            #
+            # pacman -Syu --noconfirm on an updated system:
+            # :: Synchronizing package databases...
+            #  core is up to date
+            #  extra is up to date
+            #  community is up to date
+            #  multilib is up to date
+            # :: Starting full system upgrade...
+            #  there is nothing to do  <--- needle
+            if ! pacman -Syu --noconfirm 2>&1 | grep -q "there is nothing to do" ; then
+                echo "reboot"
+            fi
+            ;;
+        *)
+            echo "WARNING: distro upgrade not supported on $SPREAD_SYSTEM"
+            ;;
+    esac
+}
